/**************************************************************************************************
 *
 *	force.c
 *
 *	part of the layout program.
 *
 *
 *	Written by Gerard Leurs, NIKHEF-H, Amsterdam.
 *	July 1991.
 *	email: a89@nikhef.nl
 *
 *************************************************************************************************/

# include "layout.h"

	/* give the coordinates and the sizes of the node */

extern int get_characteristics(hname,x,y,w,h)
char hname[MaxEnt];		/* nodename we are looking for */
int *x, *y, *w, *h;		/* x_coord, y_coord, width, heigth */
{
	struct hostentity * dummyhost2;

	dummyhost2 = host;

	while (dummyhost2 != NULL)
	{
		if (strcmp(hname,dummyhost2->hostname) == 0)
		{
			*x = dummyhost2->x_coord;	*y = dummyhost2->y_coord;
			*w = dummyhost2->width;		*h = dummyhost2->height;
			return TRUE;
		}

		dummyhost2 = dummyhost2->next;
	}

	fprintf(stderr,"get_characteristics: %s not found in hoststructure\n",hname);
	*x = *y = *w = *h = NONSENSE;
	return FALSE;
}


	/* determine the effective center of the node */

extern void nearest_point (x1, x2, w1,w2, xr1, xr2)
int x1, x2, w1, w2;		/* coordinate and size */
int *xr1, *xr2;			/* effective resulting coordinate */
{
	/* not overlapping */ 

	if (x1 + w1 <= x2)
	{
		*xr1 = x1 + w1;		*xr2 = x2;
	}
	else if (x2 + w2 <= x1)
	{
		*xr1 = x1;		*xr2 = x2 + w2;
	}

	/* overlapping, use mid point of overlapping region */

	else if (x1 <= x2)
	{
		if (x1 + w1 < x2 + w2)
			*xr1 = *xr2 = x2 + (x1 + w1 - x2) / 2;
		else
			*xr1 = *xr2 = x2 + w2 / 2;
	}
	else /* if (x1 > x2) */
	{
		if (x1 + w1 > x2 + w2)
			*xr1 = *xr2 = x1 + (x2 + w2 - x1) / 2;
		else
			*xr1 = *xr2 = x1 + w1 / 2;
	}
}


	/* determine vector between the two interacting nodes */

extern void determine_vector(hname1,hname2,vect_x,vect_y,factor)
char hname1[MaxEnt], hname2[MaxEnt];
int *vect_x, *vect_y;			/* x/y vector for hname1 */
float factor;				/* multiply the vector by this */
{
	int x_one, y_one;			/* coordinates for first node */
	int w_one, h_one;			/* sizes for first node */
	int x_two, y_two;			/* coordinates for second node */
	int w_two, h_two;			/* sizes for second node */
	int x1_center, y1_center;		/* symbolcenter node1 */
	int x2_center, y2_center;		/* symbolcenter node2 */

	/* get characteristics for the two nodes */

	if ( ! get_characteristics(hname1,&x_one,&y_one,&w_one,&h_one) )
		fprintf(stderr,"determine_vector : error in get_characteristic for %s\n",hname1);

	if ( ! get_characteristics(hname2,&x_two,&y_two,&w_two,&h_two) )
		fprintf(stderr,"determine_vector : error in get_characteristic for %s\n",hname2);

	nearest_point(x_one, x_two, w_one, w_two, &x1_center, &x2_center);
	nearest_point(y_one, y_two, h_one, h_two, &y1_center, &y2_center);

	*vect_x = (x1_center - x2_center) * factor;
	*vect_y = (y1_center - y2_center) * factor;
}


	/* determine resulting force between the two nodes
	   if linked -> attracting forces
	   if not linked -> repulsing forces */

extern void determine_force(host1,host2,force_x,force_y,linkfactor,factor)
struct hostentity * host1;		/* the node we are determining the total forces for */
struct hostentity * host2;		/* the other nodes in the nodesfile */
int *force_x, *force_y;			/* force vectors resulting from interacting hosts */
float linkfactor;			/* number_of_links / number_of_nodes */
float factor;				/* divide the vector by this factor */
{
	int vector_x, vector_y;		/* vector between centers of first and second node */
	int vect;
	int sq_vect_x,sq_vect_y;

	if (Debug)
		fprintf(stderr,"Forces of %s on %s\n",host2->hostname,host1->hostname);

	determine_vector(host1->hostname,host2->hostname,&vector_x,&vector_y,factor);

	if (linked_or_not(host1->id, host2->id))
	{
		if ( (linkfactor > 2) && (linkfactor <= 4) )
		{
			*force_x = linkfactor * -vector_x;
			*force_y = linkfactor * -vector_y;
		}
		else if (linkfactor > 4)
		{
			*force_x = 4 * -vector_x;
			*force_y = 4 * -vector_y;
		}
		else
		{
			*force_x = 2 * -vector_x;
			*force_y = 2 * -vector_y;
		}
	}
	else
	{
		sq_vect_y = vector_y * vector_y;
		if (vector_y < 0)
			sq_vect_y = - sq_vect_y;
		vect = (vector_x == 0) ? sq_vect_y : (vector_x + sq_vect_y / vector_x);
		*force_x = (vect == 0) ? (WINDOWwidth / 4) : ((WINDOWwidth / 2) / vect);

		sq_vect_x = vector_x * vector_x;
		if (vector_x < 0)
			sq_vect_x = - sq_vect_x;
		vect = (vector_y == 0) ? sq_vect_x : (sq_vect_x / vector_y + vector_y);
		*force_y = (vect == 0) ? (WINDOWheight / 4) : ((WINDOWheight / 2) / vect );
	}
}


	/* get layout changes for every node in the structure */

extern void layout_changes(linkfactor,factor)
float linkfactor;				/* number_of_links / number_of_nodes */
float factor;				/* factor to multiply/divide the forces */
{
	struct hostentity * dummyhost1;	/* host to determine the changes for */
	struct hostentity * dummyhost2;	/* the host that interacts with dummyhost1 */
	struct hostentity * newhost;
	int force_x, force_y;		/* force vectors for dummyhost1 */
	int layout_x, layout_y;		/* total of force vectors for dummyhost1 */
	int new_x;			/* new x_coordinate for the node */
	int new_y;			/* new y_coordinate for the node */

	newhost = NULL;
	dummyhost1 = host;

	while (dummyhost1 != NULL)
	{
		layout_x = layout_y = 0;
		if (Debug)
			fprintf(stderr,"layout_changes: working on %s\n",dummyhost1->hostname);

		dummyhost2 = host;

		while (dummyhost2 != NULL)
		{
			if (strcmp(dummyhost1->hostname,dummyhost2->hostname) != 0)
			{
				determine_force(dummyhost1,dummyhost2,&force_x,&force_y,linkfactor,factor);
				layout_x += force_x;
				layout_y += force_y;
			}

			dummyhost2 = dummyhost2->next;
		}

	/* divide total force_vectors by number of nodes to "prevend off-window-positioning" */

		layout_x /= number_of_nodes;
		layout_y /= number_of_nodes;

	/* determine the new coordinates for every node, using the old
	   coordinates and the grand totals of the force_vectors */

		new_x = dummyhost1->x_coord + layout_x;
		new_y = dummyhost1->y_coord + layout_y;

	/* check coordinates out of windowboundaries */

		if (new_x < boundary)
			new_x = (boundary + dummyhost1->x_coord) / 2;
		else
		{
			while ((new_x + dummyhost1->width) > (WINDOWwidth - boundary))
				new_x -= (DEFAULTwidth / 10);
		}

		if (new_y < boundary)
			new_y = (boundary + dummyhost1->y_coord) / 2;
		else
		{
			if (dummyhost1->link > 0)
			{
				while ((new_y + dummyhost1->height + ((DEFAULTheight + 5) * dummyhost1->link)) >
				       (WINDOWheight - boundary))
					new_y -= DEFAULTheight;
			}
			else
			{
				while ((new_y + dummyhost1->height) > (WINDOWheight - boundary))
					new_y -= DEFAULTheight;
			}
		}

	/* check overlap with already changed nodes */

		check_overlap_layout(newhost,dummyhost1,&new_x,&new_y,layout_x,layout_y);

		/* make coordinates for this node known in a temp. structure */

		newhost = addnodestructure(newhost,dummyhost1->id,dummyhost1->hostname,new_x,new_y,
					   dummyhost1->width,dummyhost1->height,dummyhost1->link,TRUE);

		dummyhost1 = dummyhost1->next;
	}

	/* after the complete cycle : make changes known to the original hoststructure */

	host = newhost;
}
