.TH layout 1N "17 July 1991"

.SH NAME
.TP
.B layout
\- layout a network / graph in 2D.

.SH SYNOPSIS
layout [options]

.SH DESCRIPTION
Layout creates a 2D network graph from a list of nodenames
and connections between the nodes. Nodes are drawn as
rectangular boxes with the name of the node inside the box.
.br
Connections between nodes are drawn as lines between the
boxes.
.br
Layout produces its output after a number of iterations
in which the nodes get their position in the graph. It
checks the size of the boxes, and uses these to prevent
overlapping of nodes. layout also makes sure that the
boxes stay within the window.

Layout reads from standard input, which must consist of
lines with 3 or 6 fields:
.br
The 6 field line has the following syntax:
.br
	id nodename x_coord y_coord width height.
.br
Where id is the networkidentifier of the node, nodename is
the name of the node, x_coord is the initial x_coordinate
for this node, y_coord is the initial y_coordinate for this
node, width is the initial width of the box and height is
the initial height of the box.
.br
The 3 field line has the following syntax:
.br
	interface= nodename1 nodename2.
.br
This line specifies a connection between the two nodes
nodename1 and nodename2.

Layout produces output which consists of lines with 3 or 6
fields. The 6 field line gives us the networkidentifier of
the node, the name of the node, the final x_coordinate,
the final y_coordinate, the (possibly) adjusted width of
the box and the (possibly) adjusted height of the box. The
3 field line is just a copy of the 3_field_line input.

.SH OPTIONS
.TP 8
.B -c s	
Use characterstring s for the connection-typename
between the linked nodes. Default is interface=.
.TP 8
.B -f f
Use factor f for determination of the force_vectors.
This factor is used in the determination of the
forces between the nodes. When you want a stronger
interaction between them, you have to enlarge this
factor. When you want to reduce	the interaction you
have to enter a factor between 0 and 1. Default is 1.
.TP 8
.B -h x
Define the height of the 2D window to be used for the
layout. This must be large enough to fit all nodes into
it (and move them around). See also -w. Default is 800.
.TP 8
.B -i x
Define the number of iterations. If you do not specify
the number of iterations it will be 15. A large number
of iterations (e.g. 100) is recommended.
.TP 8
.B -p
Produce postscript-output ("/tmp/psfile"). This gives a
possibility to see how the network will look like after
a number of iterations. Default is no postscript.
.TP 8
.B -r
Put the nodes at random places on the window. Nodes
with more connections to other nodes get closer to the
center of the 2D window. Default is off. However when
a node has a zero x-coordinate it gets a random
x_coordinate and when it has a zero y-coordinate it
gets a random y_coordinate.
.TP 8
.B -s
Do not check maximum boxsizes. Use this option only if
there are nodes with large sizes. The size of the box
depends primarily on the length of the name of the
node. When a node has more connections the size is
somewhat enlarged. Default is sizecheck.
.TP 8
.B -w x
Define the width of the 2D window to be used for the
layout. See also -h. Default is 1000.

.SH FILES
/tmp/psfile
	The file with postscript output of the
.br
	program.

.SH EXAMPLES
	cat links.database | layout -w 1500 -h 1000 -p
.br
In this example we use a windowwidth of 1500 and a windowheight of
1000 to layout the nodes in the file called "links.database". We also
want the results to be saved in postscript.
.br

	cat nodefile | layout -r -i 150 -p.
.br
We place the nodes in file "nodefile" at random places in a 2D
window (nodes with more links closer to the center). We determine
150 times what the interactions are between all nodes. Finally we
produce postscriptoutput.

Possible input consists of the following lines:
.br
con= ip.host:hef-router.nikhef.nl ip.network:192.16.183.0
.br
con= ip.host:hef-router.nikhef.nl ip.network:192.16.192.0
.br
con= ip.host:cwi-gw.cwi.nl ip.network:192.16.183.0
.br
con= ip.host:cwi-gw.cwi.nl ip.network:192.16.192.0
.br
ip.host:hef-router.nikhef.nl hef-router.nikhef.nl 368 432 130 13
.br
ip.network:192.16.183.0 192.16.183.0 330 394 92 12
.br
ip.network:192.16.192.0 192.16.192.0 210 464 368 19
.br
ip.host:cwi-gw.cwi.nl cwi-gw.cwi.nl 258 362 88 13

.SH DIAGNOSTICS
A large amount of time is consumed when you need a really
good layout. When the number of nodes increase the time
used for layout goes up and up.

.SH IMPLEMENTATION
Linked nodes attract, while non-linked nodes repel. The forcevector
between two nodes depents on the distance between them.
.br
Only nodes with more than one link to other nodes with more than
one link are used for the place_determination.
.br
After every 2 iterations there is a resizing of all nodes
with more links on the window (to get a nicely scaled 2D window).
After every 10 iterations there is an overlap_check.
.br
Nodes with only two connections to other nodes with more
links are placed somewhat in between these two nodes.
.br
Nodes with only one link to another node are placed close to the
node they are connected with.
.br
Nodes without connections to other nodes are placed on the bottompart
of the window (at the end of the program).
.br
Nodes with one connection to another node with only this one
connection are placed on the upperpart of the window (at the
end of the program).

.SH LIMITATIONS
The different procedures of the program (overlap checking,
force_vector determination, resizing) have not been evaluated
extensively.
.br
At this moment the number of different structures in the program make
it hard not to go nuts. It must be rewritten so that we use the
links counted for a node and not the different structures with nodes
with 0, 1 etc. links.
.br
It must be made more flexible to handle the nodes with for instance 5
or more links first, then the nodes with 4 links and so on.

.SH BUGS
The layout can be made more efficient (with a nicer spread around
the window).
.br
Time consuming with large number of nodes.

.SH AUTHOR
Gerard Leurs
.br
NIKHEF-H, Amsterdam, Netherlands

.SH "SEE ALSO"
monet(1N)
