/**************************************************************************************************
 *
 *	read.c
 *
 *	part of the layout program.
 *
 *
 *	Written by Gerard Leurs, NIKHEF-H, Amsterdam.
 *	July 1991.
 *	email: a89@nikhef.nl
 *
 *************************************************************************************************/

# include "layout.h"

	/* print usage and exit */

extern void usage(programname)
char *programname;
{
	fprintf(stderr,"Usage : %s [-i <number of iterations>] [-h height] [-w width] ",programname);
	fprintf(stderr,"[-c connectionname] [-f forcefactor] [-p] [-r] [-s]\n");
	exit (-1);
}


/* this function builds the list with all nodes, with their coordinates and sizes */
/* or with their NEW coordinates and sizes (New == TRUE) */
/* or the nodes with zero or one links (New == TRUE) */


extern struct hostentity * addnodestructure(list,id,newname,x_coord,y_coord,width,height,link,New)
struct hostentity *list;
char id[MaxEnt], newname[MaxEnt];
int x_coord, y_coord, width, height, link;
int New;			/* make temp. structure with new coordinates or not */
{
	/* if the hostname doesn't exist in the list then add a new hostname, with his coordinates and sizes */

	if (list == NULL)
	{
		list = (struct hostentity *) malloc(sizeof(struct hostentity));

		/* check if there is enough memory available */

		if (list == NULL)
		{
			fprintf(stderr,"addnode: no more memory available !\n");
			exit(3);
		}
		if (Debug)
		{
			if (New)
				fprintf(stderr,"Adding host %s to the New_nodestructure\n",newname);
			else
				fprintf(stderr,"Adding host %s to the hoststructure\n",newname);
		}
		strcpy(list->id,id);
		strcpy(list->hostname,newname);
		list->x_coord = x_coord;
		list->y_coord = y_coord;
		if (width != 0)
			list->width = width;
		else
			list->width = strlen(newname) * 7;
		if (height != 0)
			list->height = height;
		else
			list->height = DEFAULTheight;
		list->link = link;
		list->next = NULL;
		if ( ! New)
			number_of_nodes ++;
	}

	/* if the hostname already exists then print this and go on (New == FALSE) */
	/* or change coordinates (New == FALSE) */

	else if (strcmp(list->hostname,newname) == 0)
	{
		if (New)
		{
			fprintf(stderr,"addnode (new / zero or nul links): %s already exists\n",newname);
			list->x_coord = x_coord;
			list->y_coord = y_coord;
		}
		else
			fprintf(stderr,"addnode: %s already exists\n",newname);
	}

	/* go through the complete list until the end of the list
	   or until the hostname is found . Add the hostname with
	   his coordinates and sizes */

	else
		list->next = addnodestructure(list->next,id,newname,x_coord,y_coord,width,height,link,New);

	return list;
}


/* this function builds the list with all links and both connecting sites */

extern struct linkentity * addlinkstructure(linklist,typename,first_site,second_site)
struct linkentity *linklist;
char typename[MaxEnt], first_site[MaxEnt], second_site[MaxEnt];
{
	/* if the connection doesn't exist in the linklist then add a new
	connection, with its connecting sites */

	if (linklist == NULL)
	{
		linklist = (struct linkentity *) malloc(sizeof(struct linkentity));

		/* check if there is enough memory available */

		if (linklist == NULL)
		{
			fprintf(stderr,"addlink: no more memory available !\n");
			exit(3);
		}

		if (Debug)
			fprintf(stderr,"Adding %s <-> %s to the linkstructure\n",first_site,second_site);

		strcpy(linklist->linktype,typename);
		strcpy(linklist->id_one,first_site);
		strcpy(linklist->id_two,second_site);
		linklist->next = NULL;
	}

	/* if the connection already exists then print this and go on */

	else if ((strcmp(linklist->id_one,first_site) == 0)
	     &&	 (strcmp(linklist->id_two,second_site) == 0 ))
		fprintf(stderr,"addlink: %s %s - %s already exists\n",typename,first_site,second_site);

	/* go through the complete linklist until the end of the linklist
	   or until an existing connection is found .
	   Add the connection with both sites */

	else
		linklist->next = addlinkstructure(linklist->next,typename,first_site,second_site);

	return linklist;
}


	/* sort the nodestructure on length of the names */

extern void sort_the_nodestructure()
{
	struct hostentity * dummyhost1;
	struct hostentity * newhost;
	int length;

	newhost = NULL;

	for (length=1; length <= MaxEnt; length++)
	{
		dummyhost1 = host;
		while (dummyhost1 != NULL)
		{
			if ((strlen(dummyhost1->hostname)) == length)
				newhost = addnodestructure (newhost, dummyhost1->id,dummyhost1->hostname, dummyhost1->x_coord,
							   dummyhost1->y_coord, dummyhost1->width, dummyhost1->height,
							   dummyhost1->link, TRUE);
			dummyhost1 = dummyhost1->next;
		}
	}

	host = newhost;
}


	/* This function copies the built-up list of hostnames and coordinates and sizes to the output */
      
extern void printhoststructure(all)
struct hostentity *all;
{
	struct hostentity *dummy;

	dummy = all;

	while (dummy != NULL)
	{
		fprintf(stderr,"%s %s  %d  ",dummy->id,dummy->hostname,dummy->x_coord);
		fprintf(stderr,"%d  %d  %d\n",dummy->y_coord,dummy->width,dummy->height);

		dummy = dummy->next;
	}
}


	/* This function copies the built-up list of links with linked sites to the output */

extern void printlinkstructure(all)
struct linkentity *all;
{
	struct linkentity *dummy;

	dummy = all;

	while (dummy != NULL)
	{
		fprintf(stderr,"%s %s  %s\n",dummy->linktype,dummy->id_one,dummy->id_two);

		dummy = dummy->next;
	}
}
