/**************************************************************************************************
 *
 *	resize.c
 *
 *	part of the layout program.
 *
 *
 *	Written by Gerard Leurs, NIKHEF-H, Amsterdam.
 *	July 1991.
 *	email: a89@nikhef.nl
 *
 *************************************************************************************************/

# include "layout.h"

	/* determine minimum and maximum for x and y coordinates of all nodes */

static void min_and_max(x_factor,y_factor,min_x,min_y)
float *x_factor, *y_factor;	/* new scaling factors */
int *min_x, *min_y;		/* minimum found x and y coordinate */
{
	struct hostentity * dummyhost1;	/* host to determine the changes for */
	int max_x, max_y;		/* maximum x/y_coordinate for all nodes */

	if (Debug)
		fprintf(stderr,"getting minimum and maximum x and y coordinates\n");

	*min_x = WINDOWwidth;		max_x = 0;
	*min_y = WINDOWheight;   	max_y = 0;

	dummyhost1 = host;

	while (dummyhost1 != NULL)
	{
		if (dummyhost1->x_coord < *min_x)
			*min_x = dummyhost1->x_coord;
		if ((dummyhost1->x_coord + dummyhost1->width) > max_x)
			max_x = dummyhost1->x_coord + dummyhost1->width;
		if (dummyhost1->y_coord < *min_y)
			*min_y = dummyhost1->y_coord;
		if (dummyhost1->link > 0)
		{
			if ((dummyhost1->y_coord + dummyhost1->height + (DEFAULTheight * (1 + dummyhost1->link))) > max_y)
				max_y = dummyhost1->y_coord + dummyhost1->height + (DEFAULTheight * (1 + dummyhost1->link));
		}
		else
		{
			if ((dummyhost1->y_coord + (2 * dummyhost1->height)) > max_y)
				max_y = dummyhost1->y_coord + (2 * dummyhost1->height);
		}

		dummyhost1 = dummyhost1->next;
	}

	/* boundary is used for unlinked nodes so we must have a clean surrounding area */

	if (*min_x < boundary)
		*min_x = boundary;
	if (max_x > WINDOWwidth)
		max_x = WINDOWwidth;
	if (*min_y < boundary)
		*min_y = boundary;
	if (max_y > WINDOWheight)
		max_y = WINDOWheight;

	if (max_x != *min_x)
		*x_factor = (float) (WINDOWwidth  - 2 * boundary) / (max_x - *min_x);
	else
		*x_factor = 1.0;
	if (max_y != *min_y)
		*y_factor = (float) (WINDOWheight - 2 * boundary) / (max_y - *min_y);
	else
		*y_factor = 1.0;

	if (Debug)
	{
		fprintf(stderr,"scaling factors are : %f %f\n",*x_factor,*y_factor);
		fprintf(stderr,"minimum x / y coordinates are : %d %d\n",*min_x,*min_y);
	}
}


	/* move nodes in the adjacent sectors towards the emptiest sector */

static void sector_adjusting(sectornr,scale_x,scale_y)
int sectornr;		/* the emptiest sector of the window */
float scale_x, scale_y;	/* the scale factors (x and y) for the complete window */
{
	struct hostentity * dummyhost1;		/* host to determine the changes for */
	int min_x, min_y;			/* x and y coordinates for the empty sector */
	int max_x, max_y;			/* x and y coordinates for the empty sector */
	int x, y, w, h;				/* temp. coordinates and sizes */
	int new_x, new_y;			/* temp. new coordinates for moved nodes */
	float move_factor_x, move_factor_y;	/* factor to move adjacent sector-nodes */
	int nodes_moved = 0;			/* nodes moved towards the emptiest sector */

	if (sectornr == 1)
	{
		min_x = boundary;		min_y = WINDOWheight / 2;
		max_x = WINDOWwidth / 2;	max_y = WINDOWheight - boundary;
	}
	else if (sectornr == 2)
	{
		min_x = WINDOWwidth / 2;	min_y = WINDOWheight / 2;
		max_x = WINDOWwidth - boundary;	max_y = WINDOWheight - boundary;
	}
	else if (sectornr == 3)
	{
		min_x = boundary;		min_y = boundary;
		max_x = WINDOWwidth / 2;	max_y = WINDOWheight / 2;
	}
	else
	{
		min_x = WINDOWwidth / 2;	min_y = boundary;
		max_x = WINDOWwidth - boundary;	max_y = WINDOWheight / 2;
	}

	/* get minimum scale factor (x and y) */

	if (scale_x <= 1.2)
		move_factor_x = 1.2;
	else
		move_factor_x = scale_x;
	if (scale_y <= 1.2)
		move_factor_y = 1.2;
	else
		move_factor_y = scale_y;

	/* give the nodes in adjacent sectors new coordinates */

	dummyhost1 = host;

	while (dummyhost1 != NULL)
	{
		new_x = x = dummyhost1->x_coord;	new_y = y = dummyhost1->y_coord;
		w = dummyhost1->width;
		if (dummyhost1->link > 0)
			h = dummyhost1->height + (DEFAULTheight * (1 + dummyhost1->link));
		else
			h = dummyhost1->height;

		/* determine in which sector the nodes are and move them */

		if ((x >= min_x) && (x <= max_x))
		{
			if (y < min_y)
			{
				if (Debug)
					fprintf(stderr,"moving (y_direction) %s towards sector %d\n",dummyhost1->hostname,sectornr);

				new_y = dummyhost1->y_coord * move_factor_y;
				nodes_moved ++;
			}
			else if (y > max_y)
			{
				if (Debug)
					fprintf(stderr,"moving (y_direction)%s towards sector %d\n",dummyhost1->hostname,sectornr);

				new_y = dummyhost1->y_coord / move_factor_y;
				nodes_moved ++;
			}
		}
		else
		{
			/* place the nodes of the non-adjacent sector slightly towards the two adjacent sectors */

			if (Debug)
				fprintf(stderr,"slightly adjusting %s towards adjacent sector\n",dummyhost1->hostname);

			if (y < (min_y / 2))
				new_y = dummyhost1->y_coord * 1.1;
			else if (y > (max_y * 1.5))
				new_y = dummyhost1->y_coord / 1.1;
		}

		if ((y >= min_y) && (y <= max_y))
		{
			if (x < min_x)
			{
				if (Debug)
					fprintf(stderr,"moving (x_direction) %s towards sector %d\n",dummyhost1->hostname,sectornr);

				new_x = dummyhost1->x_coord * move_factor_x;
				nodes_moved ++;
			}
			else if (x > max_x)
			{
				if (Debug)
					fprintf(stderr,"moving (x_direction) %s towards sector %d\n",dummyhost1->hostname,sectornr);

				new_x = dummyhost1->x_coord / move_factor_x;
				nodes_moved ++;
			}
		}
		else
		{
			/* place the nodes of the non-adjacent sector slightly towards the two adjacent sectors */

			if (Debug)
				fprintf(stderr,"slightly adjusting %s towards adjacent sector\n",dummyhost1->hostname);

			if (x < (min_x / 2))
				new_x = dummyhost1->x_coord * 1.2;
			else if (x > (max_x * 1.5))
				new_x = dummyhost1->x_coord / 1.2;
		}

		/* check boundaries */

		if (new_x < boundary)
			new_x = boundary;
		if ((new_x + w) > (WINDOWwidth - boundary))
			new_x = WINDOWwidth - boundary - w;

		if (new_y < boundary)
			new_y = boundary;
		if ((new_y + h) > (WINDOWheight - boundary))
			new_y = WINDOWheight - boundary - h;

		dummyhost1->x_coord = new_x;
		dummyhost1->y_coord = new_y;

		dummyhost1 = dummyhost1->next;
	}

	fprintf(stderr,"number of nodes moved towards sector %d is %d\n", sectornr, nodes_moved);
}


	/* after a few iterations there tends to be a shift for all nodes to
	   one direction. So we need to adjust the coordinates for all nodes
	   in order to get them nicely spread on the window */
	/* a second feature is a sector check : which sector is the most empty
	   one ? This one attracts nodes from the two adjacent sectors */

extern void resizing(sectorcheck, adjustcheck)
int sectorcheck, adjustcheck;
{
	struct hostentity * dummyhost1;		/* host to determine the changes for */
	int x1, y1, w1, h1;			/* temp. values for host1 */
	int new_x;				/* new x_coordinate for the node */
	int new_y;				/* new y_coordinate for the node */
	float scale_x, scale_y;			/* scale factor for new layout */
	int min_x, min_y;			/* minimum x and y coordinates for all nodes */
	int sector_tl, sector_tr, sector_bl, sector_br;		/* number of nodes in a kwadrant */
	int empty_sector;


	if (sectorcheck)
		fprintf(stderr,"Resizing nodes_with_more_links (with sectorcheck)\n");
  	if (adjustcheck)
  		fprintf(stderr,"Resizing nodes and rearranging of some of these nodes\n");

	/* place the nodes with only one_link (at this moment !!) close to the more_linked_node */
	/* place the nodes with two links to more_links_nodes in between these two nodes */

	place_node_1_and_2(adjustcheck);

	/* get scaling factors for all nodes (whole window) */

	min_and_max(&scale_x, &scale_y, &min_x, &min_y);

	/* give the nodes new coordinates */

	sector_tl = sector_tr = sector_bl = sector_br = 0;

	dummyhost1 = host;

	while (dummyhost1 != NULL)
	{
		if (Debug)
			fprintf(stderr,"scaling on %s\n",dummyhost1->hostname);

		x1 = dummyhost1->x_coord;	y1 = dummyhost1->y_coord;
		w1 = dummyhost1->width;
		if (dummyhost1->link > 0)
			h1 = dummyhost1->height + (DEFAULTheight * (1 + dummyhost1->link));
		else
			h1 = dummyhost1->height * 2;

		new_x = boundary + ((x1 - min_x) * scale_x);
		if (new_x < boundary)
			new_x = boundary;
		if ((new_x + w1) > (WINDOWwidth - boundary))
			new_x = WINDOWwidth - boundary - w1;

		new_y = boundary + ((y1 - min_y) * scale_y);
		if (new_y < boundary)
			new_y = boundary;
		if ((new_y + h1) > (WINDOWheight - boundary))
			new_y = WINDOWheight - boundary - h1;
		dummyhost1->x_coord = new_x;
		dummyhost1->y_coord = new_y;

		/* determine in which sector the nodes are : bottomleft, bottomright, topleft or topright */

		if ( (dummyhost1->x_coord + dummyhost1->width / 2) <= (WINDOWwidth / 2) )
		{
			if ( (dummyhost1->y_coord + dummyhost1->height / 2) > (WINDOWheight / 2) )
				sector_tl ++;
			else
				sector_bl ++;
		}
		else
		{
			if ( (dummyhost1->y_coord + dummyhost1->height / 2) > (WINDOWheight / 2) )
				sector_tr ++;
			else
				sector_br ++;
		}

		dummyhost1 = dummyhost1->next;
	}

	if (Debug)
		fprintf(stderr,"number of nodes in sectors 1, 2, 3, 4 are: %d %d %d %d\n",sector_tl,sector_tr,sector_bl,sector_br);

	/* get the sector with lowest number of nodes. Don't do anything if two sectors are equally empty */
	/* nodes of the two adjacent kwadrants are moved towards the emptiest kwadrant */
	/* make an empty corner in the emptiest sector so we can use it
	   to place the more_linked_nodes with only connections to one_link_nodes */

	if (sectorcheck)
	{
		if (sector_tl < sector_tr)
		{
			if (sector_bl < sector_br)
			{
				if (sector_tl < sector_bl)
					sector_adjusting(1,scale_x,scale_y);
				else if (sector_tl > sector_bl)
					sector_adjusting(3,scale_x,scale_y);
			}
			else if (sector_bl > sector_br)
			{
				if (sector_tl < sector_br)
					sector_adjusting(1,scale_x,scale_y);
				else if (sector_tl > sector_br)
					sector_adjusting(4,scale_x,scale_y);
			}
		}
		else if (sector_tl > sector_tr)
		{
			if (sector_bl < sector_br)
			{
				if (sector_tr < sector_bl)
					sector_adjusting(2,scale_x,scale_y);
				else if (sector_tr > sector_bl)
					sector_adjusting(3,scale_x,scale_y);
			}
			else if (sector_bl > sector_br)
			{
				if (sector_tr < sector_br)
					sector_adjusting(2,scale_x,scale_y);
				else if (sector_tr > sector_br)
					sector_adjusting(4,scale_x,scale_y);
			}
		}
	}
}
