/******************************************************************/
/******************************************************************/
/*  $Header$  */
/******************************************************************/
/*  This is part of the user interface and handles much of the    */
/*  X windows stuff.  A map area is the collection of X wigdets   */
/*  used to display a map on the screen.                          */
/******************************************************************/

#include "xmonet.h"

/******************************************************************/
/*  Globals  */
/******************************************************************/

static Cursor    NormCursor, SizingCursor, WaitCursor, ExecCursor;

#define MAPACTION(name, params_required) \
	ACTION(name, params_required) \
	Map map; \
	if (!(map = widget2map (widget))) { \
		sys_error ("map is null"); \
	} else { CurrentMap = map; {
#define ENDMAPACT ENDACT }}

/****************************************************************************/
/*  Get position of cursor in a widget  */
/*  !!! fixforscale */
/****************************************************************************/

extern Boolean create_new_drawing_at_cursor (class, title)
	Entity class;
	String title;
{
	Widget widget;
	Window junk1, junk2;
	int root_x, root_y, win_y, win_x;
	unsigned int junkM;
	Position xloc, yloc, x,y;
	Dimension width, height;
	Map map;
	Drawing new;

	map = Map_root (CurrentMap);
	if (map == NULL) {
		user_error ("You must select a map first");
		return False;
	} 
	widget = Map_widget (map);

	if (widget == NULL)
		return False;

	/* The width will not be correct unless it is realized. */
	if (!XtIsRealized(widget))
		XtRealizeWidget(widget);
	
	if (XQueryPointer(XtDisplay(widget), XtWindow(widget), &junk1, &junk2, 
		&root_x, &root_y, &win_x, &win_y, &junkM) == False) {
		user_error ("Could not find location of mouse pointer");
		return False;
	}
	x = win_x;
	y = win_y;
	(void) Map_get_marked (&map, &x, &y, &xloc, &yloc, &width, &height);
	Map_adjust_for_grid (map, x, y, &x, &y);

	if (Map_can_insert_drawing (map, class) == False) {
		user_error ("Cannot insert instances of class [%s] in map [%s]",
			Entity_title (class), Map_title (map));
	} else if (! is_title_valid (class, title)) {
		user_error ("Title [%s] does not match the pattern [%s] for class [%s]",
			title, get_titleFormat (class), Entity_title (class));
	} else if (new = Drawing_create (title, class, map, x,y, 0,0)) {
		free_selection ();
		user_inform ("New drawing [%s] created, now position it", Drawing_title (new));
		add_selection (new);
		return True;
	} else {
		user_error ("Could not create drawing");
	}
	return False;
}

/*****************************************************************************/
/*  Takes a widget and finds the associated map area */
/*****************************************************************************/

static Map widget2map(w)
	Widget w;
{
	Map m;

	FORALLMAPS (m)
		if (w == m->mapWidget)
			return m;
	sys_error ("Map not found from widget %d", w);
	return NULL;
}

/******************************************************************/
/*  Cursor stuff                                                  */
/******************************************************************/

static void map_cursor (cursor)
	Cursor cursor;
{
	Map m;
	FORALLMAPS (m)
		if (m->mapWidget)
			XDefineCursor (XtDisplay(m->mapWidget), XtWindow(m->mapWidget), cursor);
}

extern void Map_sizing_cursor ()
{ map_cursor (SizingCursor); }

extern void Map_normal_cursor ()
{ map_cursor (NormCursor); }

extern void Map_wait_cursor ()
{ map_cursor (WaitCursor); }

extern void Map_exec_cursor ()
{ map_cursor (ExecCursor); }

/*****************************************************************************/
/* Adjust size of map area to be just large enough for the objects in it     */
/* !!! Does not handle 'scaled' maps, only default */
/*****************************************************************************/

extern void	Scale2ActualSize PR((Scale, Dimension, Dimension, Dimension*, Dimension*));
static void Map_adjust_size (map)
	Map map;
{
	Position xl, yl, xr, yr, x,y;
	Dimension w, h, w_popup, h_popup;

	if ( ! map->mapWidget) {
		if (map->parent) {
			Drawing_unmoved_undraw (map->parent);
			Drawing_unmoved_redraw (map->parent);
			return;
		}
	}

	if (map->mapWidget == NULL) {
		user_warning ("Map widget is null");
		return;
	}
	if (Map_corners (map, &xl,&yl, &xr,&yr, False)) {
		Widget_get_dimensions (XtParent(map->mapWidget), &x,&y,&w_popup,&h_popup);
		w = xr;
		h = yr;
		if (w<w_popup)
			w = w_popup;
		if (h<h_popup)
			h = h_popup;
		if (w > 0 && h > 0) {
			user_inform ("Adjusting size to [%d,%d]", w, h);
			Scale2ActualSize (map->scale, w, h, & w, & h);
			w += MAP_MARGIN_SIZE * 2;
			h += MAP_MARGIN_SIZE * 2;
			if (XtGeometryAlmost == XtMakeResizeRequest(map->mapWidget, w, h, &w, &h)) {
				/* If can't get exact values then make do with those given */
				(void) XtMakeResizeRequest(map->mapWidget, w, h, (Dimension *) NULL, (Dimension *) NULL);
			}
		}
	}
}

/******************************************************************/
/*  Should use this to redraw X11 maps.  This clears the whole    */
/*  then sends an Expose event.  Does the redraw Asyncronously.   */
/*  SHould be clever if map has no widget.                        */
/******************************************************************/

extern void Map_request_redraw (map)
	Map map;
{
	Widget widget;
	Window win;
	XEvent event;

	map = Map_root(map);
	widget = map->mapWidget;

	if (XtIsRealized (widget) == False)
		return;
	if (isDrawingAllowed () == False)
		return;

	win  = XtWindow (widget);
	while (XCheckWindowEvent (display, win, Expose, &event)) {
		/*  Remove any expose events already in the queue, no  */
		/*  need to repeat ourselves.  */
	}
	/*  Clears the whole map widget and sends an expose event  */
	if (map->drawnas == drawnas_list) {
		setup_list_map (map);
		Map_adjust_size (map);
	}
	XClearArea (display, win, 0,0 ,0,0, True);
}

/******************************************************************/
/*  Extra Annotation  */
/******************************************************************/

static void draw_page (map)
	Map map;
{
	Dimension w, h;
	Position x, y;

	if (map->mapWidget == NULL)
		return;
	Widget_get_dimensions (map->mapWidget, &x, &y, &w, &h);
	Draw_draw_rectangle(shade_grey,
		MAP_MARGIN_SIZE,         MAP_MARGIN_SIZE,
		w - MAP_MARGIN_SIZE * 2, h - MAP_MARGIN_SIZE * 2);
	if (map->marked) {
		Draw_draw_rectangle(shade_dark,
			MAP_MARGIN_SIZE+1,           MAP_MARGIN_SIZE + 1,
			w - MAP_MARGIN_SIZE * 2 - 2, h - MAP_MARGIN_SIZE * 2 - 2);
	}
}


/******************************************************************/
/*  Part of the map has been exposed  */
/*  All map redraws should send an expose event and not try to */
/*  redraw the window syncronously  */
/******************************************************************/

MAPACTION(Map_redrawACT, 0)
	XRectangle rectangle;

	if (event->type == Expose) {
		/*  Could replace this lot with the one statement :
		    XtAddExposureToRegion (event, map->clip_region); */
		/***	do { */
		/*  Search through the event list for all Expose  */
		/*  events.  Best deal with 'everything' at once. */
		/*  Merge the events into a clip-region before    */
		/*  redrawing the map.                            */
		/*  !!! May not work. */
		rectangle.x = (Position) event->xexpose.x;
		rectangle.y = (Position) event->xexpose.y;
		rectangle.width  = (Dimension) event->xexpose.width;
		rectangle.height = (Dimension) event->xexpose.height;
		XUnionRectWithRegion (&rectangle, map->clip_region, map->clip_region);

		XClearArea (display, XtWindow(widget),
			rectangle.x, rectangle.y,
			rectangle.width, rectangle.height,
			False);

		/***	} while (XCheckWindowEvent (display, win, Expose, event)); */
		if (event->xexpose.count == 0 && isDrawingAllowed ()) {
			Map_wait_cursor ();
			if (start_draw_X11 (Map_widget(map), map->scale, map->clip_region, True)) {
				draw_page (map);
				Map_draw (map);
                		XDestroyRegion (map->clip_region);
				map->clip_region = XCreateRegion ();
			}
			Map_normal_cursor ();
		}		
	} else {
		/* -- This is not an expose event, so redraw the whole lot */
		/* -- This may be inefficient, but it should be more robust */
		Map_request_redraw (map);
	}
ENDMAPACT

/******************************************************************/
/*  The cursor has moved into the map area  */
/******************************************************************/

MAPACTION(Map_selectedACT, 0)
	CurrentMap = map;
ENDMAPACT

MAPACTION(Map_unselectedACT, 0)
ENDMAPACT

/*****************************************************************/
/*  Resize a drawing marked by the cursor  */
/*****************************************************************/

MAPACTION(MapResizeDrawingACT, 0)
	Position xloc, yloc, x,y;
	Dimension width, height;
	Drawing s;
	Map cursor_map = map, drawing_map;

	x = event->xbutton.x;
	y = event->xbutton.y;

	if (map->drawnas == drawnas_map &&
		(s = Map_get_marked (&cursor_map, &x, &y, &xloc, &yloc, &width, &height))) {
		if (Drawing_isResizable (s)) {
			drawing_map = Drawing_mapOf (s);
			Drawing_undraw (s);
			Map_sizing_cursor ();
			if (Xio_rubberband_makebox (widget, &xloc, &yloc, &width, &height, *event)) {

				Position xl, yl, xr, yr;

				Drawing_redraw (s);
				Map_actual2map (&map, &xloc, &yloc, &width, &height);
				Map_corners (map, &xl, &yl, &xr, &yr, False);
/*				if (drawing_map != map) {
					user_error ("Cannot resize out of the map");
				} else if (xloc < xl || yloc < yl || xloc+width > xr || yloc+height > yr) {
					user_error ("Can't resize bigger than map");
				} else */
				{
					Map_insert_drawing (map, s, xloc, yloc, width, height);
				}
			} else
				Drawing_redraw (s);
			Map_normal_cursor ();
		} else {
			user_error ("[%s] is not resizable", Drawing_title(s));
		}
	}
ENDMAPACT

MAPACTION(MapSelectDrawingACT, 0)
	Position xloc, yloc;
	Dimension width, height;
	Drawing s;

	xloc = event->xbutton.x;
	yloc = event->xbutton.y;
	width = 0;
	height = 0;

	Map_sizing_cursor ();
	(void) Xio_rubberband_makebox (widget, &xloc, &yloc, &width, &height, *event);
	if (Map_select_in_box (map, xloc, yloc, width, height) == False)
		free_selection ();
	Map_normal_cursor ();

ENDMAPACT

/*****************************************************************/
/*  Drag one or more drawings across the map  */
/*****************************************************************/

static Drawing SingleSelected;
static Map MoveMap;

#define dragable(s) (Drawing_mapOf(s)==MoveMap && Drawing_isMoveable(s))


/*  !!! fixforscale */

static void drag_selected (x0, y0, x,y)
	Position x,y, x0, y0;
{
	Drawing d;
	Map map;
	Position cx = x - x0, cy = y - y0;

	map = Map_root(CurrentMap);
	(void) Map_get_marked (&map, &x0, &y0,  NULL, NULL, NULL, NULL);
	map = Map_root(CurrentMap);
	(void) Map_get_marked (&map, &x, &y,  NULL, NULL, NULL, NULL);
	FOR_SELECTIONS (d)
		if (dragable(d))
			Drawing_outline (d, map, X(d)+x-x0,Y(d)+y-y0);
}

static void draw_selected (x0, y0, x,y)
	Position x,y, x0, y0;
{
	Dimension width, height;
	Drawing d;
	Map map;

	map = Map_root(CurrentMap);
	(void) Map_get_marked (&map, &x0, &y0,  NULL, NULL, NULL, NULL);
	map = Map_root(CurrentMap);
	(void) Map_get_marked (&map, &x, &y, NULL, NULL, NULL, NULL);
	FOR_SELECTIONS (d)
		if (dragable(d))
			Map_insert_drawing (map, d, X(d)+x-x0, Y(d)+y-y0, (Dimension) 0, (Dimension) 0);
}

static void drag_single_selected (x0, y0, x,y)
	Position x,y, x0, y0;
{
	Dimension width, height;
	Map map;
	Position cx = x - x0, cy = y - y0;

	map = CurrentMap;
	(void) Map_get_marked (&map, &x0, &y0,  NULL, NULL, NULL, NULL);
	map = CurrentMap;
	(void) Map_get_marked (&map, &x, &y,  NULL, NULL, NULL, NULL);
	Drawing_outline (SingleSelected, map, X(SingleSelected)+x-x0, Y(SingleSelected)+y-y0);
}

/*  !!! fixforscale */

static void draw_single_selected (x0, y0, x,y)
	Position x,y, x0, y0;
{
	Dimension width, height;
	Map map;
	Position cx = x - x0, cy = y - y0;

	map = CurrentMap;
	(void) Map_get_marked (&map, &x0, &y0,  NULL, NULL, NULL, NULL);
	map = CurrentMap;
	(void) Map_get_marked (&map, &x, &y,  NULL, NULL, NULL, NULL);
	Map_insert_drawing (map, SingleSelected,
		X(SingleSelected)+x-x0, Y(SingleSelected)+y-y0,
		 (Dimension) 0, (Dimension) 0);
}

MAPACTION(MapMoveDrawingACT, 0)
	Drawing s;
	Position xloc, yloc;
	Dimension width, height;

	if (map->drawnas != drawnas_map)
		return;

	xloc = event->xbutton.x;
	yloc = event->xbutton.y;

	if ((s = Map_get_marked (&map, &xloc, &yloc, NULL, NULL, NULL, NULL)) &&
		!(s->submap && current_Selection ())) {
		if (Drawing_isMoveable (s)) {
			user_inform ("Moving [%s]", Drawing_title (s));
			Map_sizing_cursor ();
			SingleSelected = s;
			MoveMap = map;
			(void) start_draw_X11 (Map_widget(map), map->scale, NULL, False);
			Xio_rubberband_drag (
				widget, map->grid_x, map->grid_y,
				drag_single_selected, draw_single_selected, *event);
			Draw_end_draw ();
			Map_normal_cursor ();
		} else {
			user_error ("System [%s] is not moveable", Drawing_title (s));
		}
	} else {
		MoveMap = map;
		(void) start_draw_X11 (Map_widget(map), map->scale, NULL,  False);
		Xio_rubberband_drag (
			widget, map->grid_x, map->grid_y,
			drag_selected, draw_selected, *event);
		Draw_end_draw ();
	}
ENDMAPACT

/*****************************************************************/
/*  Perform an action on drawing the cursor is currently above   */
/*****************************************************************/

MAPACTION(Map_drawingActionACT, 0)
	Drawing s;
	Position xloc, yloc;

	xloc = event->xbutton.x;
	yloc = event->xbutton.y;

	if (s = Map_get_marked (&map, &xloc, &yloc, NULL, NULL, NULL, NULL)) {
		Drawing_do_action (s);
	} else {
		user_inform ("Freeing Selected Objects");
		free_selection ();
	}
ENDMAPACT

/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/
/*****************************************************************************/

/*****************************************************************************/
/*   Has a list of all defined maps, for the user to open or close           */
/*****************************************************************************/

CALLBACK(Map_CornersCB)
	Map_adjust_size (Map_root(CurrentMap));
ENDCB

ACTION(Map_cornersACT, 0)
	Map_CornersCB (widget, NULL, NULL);
ENDACT

/******************************************************************/
/*  Callbacks  */
/******************************************************************/

CALLBACK(Map_freeCB)
	if (sizeOf_drawingList(CurrentMap->elements) > 0) {
		sys_error("Can't free a map with sub components");
	} else {
		XtFree (CurrentMap->clip_region);
		CurrentMap->mapWidget = NULL;
		Map_free (CurrentMap);
	}
ENDCB

CALLBACK(Map_redrawCB)
	Map_request_redraw (CurrentMap);
ENDCB

CALLBACK(Map_createCB)
	user_error ("Creating map for widget");
	CurrentMap = Map_create ();
	CurrentMap -> mapWidget = widget;
	CurrentMap -> clip_region = XCreateRegion ();
ENDCB

/******************************************************************/
/*  This routine is called every time the size of a map is        */
/*  changed.                                                      */
/******************************************************************/

extern void Map_changed_handler (map)
	Map map;
{
	Map_adjust_size (map);
}

/*****************************************************************************/
/* ??? */
/*****************************************************************************/

static Widget    MapParentWidget;

/******************************************************************/
/*  Initialise and finish off                                     */
/******************************************************************/

extern void MapCB_init ()
{
#define NUM_ACTIONS 8
	static XtActionsRec actions[NUM_ACTIONS] =
	{
	 {"MapMoveDrawingACT",		MapMoveDrawingACT},
	 {"MapSelectDrawingACT",	MapSelectDrawingACT},
	 {"MapResizeDrawingACT",	MapResizeDrawingACT},
	 {"MapSelectedACT",		Map_selectedACT},
	 {"MapUnselectedACT",		Map_unselectedACT},
	 {"MapActionACT",		Map_drawingActionACT},
	 {"MapRedrawACT",		Map_redrawACT},
	 {"MapCornersACT",		Map_cornersACT}
	};

	init_frontend();

	(void) XtAppAddActions(app_con, (XtActionList) actions, (Cardinal) NUM_ACTIONS);

	/*  Set up default callbacks  */
	RCALL("MapFreeCB",		Map_freeCB,		NULL);
	RCALL("MapCreateCB",		Map_createCB,		NULL);
	RCALL("MapRedrawCB",		Map_redrawCB,		NULL);
	RCALL("MapCornersCB",		Map_CornersCB,		NULL);
	/*  Set up cursors  */
	NormCursor   = XCreateFontCursor(display, XC_crosshair);
	SizingCursor = XCreateFontCursor(display, XC_sizing);
	WaitCursor   = XCreateFontCursor(display, XC_watch);
	ExecCursor   = XCreateFontCursor(display, XC_leftbutton);

	MapParentWidget = toplevel;
	CurrentMap = MarkedMap = NULL;
}

extern void MapCB_finish ()
{
	MapParentWidget = (Widget) 0;
}

/******************************************************************/
/******************************************************************/
/*  $Header$  */
/******************************************************************/
/******************************************************************/


