#ifndef  MONET_ENTITY_H
#define  MONET_ENTITY_H

/***************************************************************************/
/***************************************************************************/
/*  $Header$ */
/***************************************************************************/
/***************************************************************************/

#include "general/General.h"
#include "general/SimpleList.h"
#include "general/Shape.h"
typedef SimpleList EntityList;

typedef struct protocol_entity	* Entity;

typedef struct drawing		* Drawing ;
typedef struct request		* Request;
typedef struct attribute	* Attribute;
typedef struct map		* Map;

/***************************************************************************/
/*  Constants                                                              */
/***************************************************************************/

#define MAXCONNECTIONS		-1   /* !!! */
#define ENTITY_MAXPATHLEN   1024
#define ENTITY_MAXNAMELEN   64

/***************************************************************************/
/*  Forward definition of some private datastructures                      */
/***************************************************************************/

typedef struct query_protocol	* QueryClass;
typedef struct query	* Query;
typedef unsigned long		TTime;

/***************************************************************************/
/*  What kind of visual subclass                                           */
/***************************************************************************/

typedef enum {
	o_none,
	o_interface,
	o_map,
	o_system,
	} VInstanceType;

/***************************************************************************/
/*  Includes                                                               */
/***************************************************************************/

#include "db/Status.h"  /*  This will also be merged later  */
#include "db/Attribute.h"
#include "db/Interface.h"
#include "db/Component.h"
#include "db/Class.h"
#include "db/Redirect.h"

/***************************************************************************/
/*  Lists of entities, these are created from a generic list module.       */
/***************************************************************************/

#define NULL_OL  (EntityList) 0

#define extract_entityList(list,o)	extract_simpleList (list, (caddr_t) o)
#define create_entityList()		(EntityList) create_simpleList ()
#define reset_entityList(list)		reset_simpleList (list)
#define free_entityList(list)		free_simpleList (list)
#define insert_entityList(list, o)	insert_simpleList(list, (caddr_t) o)
#define isEmpty_entityList(list)	isEmpty_simpleList(list)
#define getN_entityList(list, n)	(Entity) getN_simpleList(list,n)
#define first_entityList(list)		(Entity) first_simpleList(list)
#define next_entityList(list)		(Entity) next_simpleList(list)
#define sizeOf_entityList(list) 	sizeOf_simpleList(list)
#define second_entityList(list)		(Entity) second_simpleList(list)
#define isMember_entityList(list, o)	isMember_simpleList(list, (caddr_t) o)

#define FOR_ENTITYLIST(o, list) for (o=first_entityList (list); o != NULL; o=next_entityList (list))

/***************************************************************************/
/*  A protocol Entity                                                      */
/*  !!! This has been formed by merging two different structs together     */
/*  !!! that had a fair bit in common.                                     */
/***************************************************************************/

struct protocol_entity {

	/***************************************************************************/
	/*  Definition of a visual object.                                         */
	/***************************************************************************/

	Entity		class;		/* Indicates superclass */
#define Entity_classOf(o)	(o == NULL ? NULL : o->class)

	VInstanceType	instance_type;	/* Indicate base type of instance */
#define Entity_typeOf(o) (o->instance_type)

	Boolean		standard;	/* This object will not be saved */

	/***************************************************************************/
	/*  Objects referencing the entity.  Currently two types, each is          */
	/*  (or should be) a list of objects all referencing this particular       */
	/*  entity.  !!! Should generalise this so that no changes need be made    */
	/*  here for new types.                                                    */
	/***************************************************************************/

	Request		requests;	/* List of requests to this entity */
	Map		map;
	Drawing         drawing;	/* List of visual objects for this entity */
	Shape		shape;
	Query		query;

	/***************************************************************************/
	/*  Special stuff for entities that are classes                            */
	/***************************************************************************/

	EntityClassData	class_stuff;	/*  For Classes Only */
#define Entity_isClass(o)	(o->class_stuff != NULL)

	/***************************************************************************/
	/*  The namespace heirarchy.  Entities are given a domain name and are thus*/
	/*  orgainised in a heirarchy.   */
	/***************************************************************************/

	/* !!! Not yet Implemented */
	Entity parent, child, next_sibling;

	/***************************************************************************/
	/*  The following defines the entities status                              */
	/*  The entity database is described as a stack of networks.  Each entity  */
	/*  is either a node/system or a link/interface.  A node is connected to   */
	/*  other nodes via a set of interfaces, while an interface connects at    */
	/*  at most two systems together.                                          */
	/***************************************************************************/

	Entity		from, to;	/* For Interfaces or Interface Classes Only */
#define FROM(o)				(o->from)
#define TO(o)				(o->to)

	EntityList	interfaces;	/* For Systems Only */
#define Entity_interfaces(o)		(o->interfaces)
#define INTERFACES(o)			(o->interfaces)

	EntityList	supercomponents;
#define Entity_supercomponents(o)	(o->supercomponents)

	EntityList	subcomponents;
#define Entity_subclasses(o)		(o->subcomponents)
#define Entity_subcomponents(o)	(o->subcomponents)

	VStatus		status;
#define V_STATUS(o)		(o->status)
#define PINGID(o)		(V_STATUS(o)->ping_id)
#define ALIAS(o)		(V_STATUS(o)->alias)
#define STATUS(o)		(V_STATUS(o)->status)
#define TITLE(o)		(V_STATUS(o)->title)
#define MASK(o)			(V_STATUS(o)->mask)
#define FUTURE(o)		(V_STATUS(o)->future)
#define CAPACITY(o)		(V_STATUS(o)->capacity)
#define CONNECTIONS(o)		(V_STATUS(o)->num_connections)
#define PING(o)			(V_STATUS(o)->ping)
#define BEHAVIOUR(o)		(V_STATUS(o)->behaviour)
#define LOADIN(o)		(V_STATUS(o)->load_in)
#define LOADOUT(o)		(V_STATUS(o)->load_out)
#define PKTIN(o)		(V_STATUS(o)->pkts_in)
#define PKTOUT(o)		(V_STATUS(o)->pkts_out)
#define OCTIN(o)		(V_STATUS(o)->octets_in)
#define OCTOUT(o)		(V_STATUS(o)->octets_out)
#define CONTACT(o)		(V_STATUS(o)->contact)
#define ENTACTION(o)		(V_STATUS(o)->action)

	Attribute	attributes;	/* List of attributes associated with this object */

	/***************************************************************************/
	/*  Where to send query requests to                                        */
	/***************************************************************************/

	RedirectRule	redirect;

	/***************************************************************************/
	/* To ensure correct allocation */
	/***************************************************************************/

	int num_lists;  /* Number of object lists this is in */

/***************************************************************************/
};  /*****  End of struct protocol_entity ****/
/***************************************************************************/

extern Entity TopClass;

/***************************************************************************/
/*  Function Prototypes                                                    */
/***************************************************************************/

#ifdef MONET_PROTOTYPES

extern void		ClassEntity_init		(void);
extern void		ClassEntity_finish		(void);
extern Boolean		ClassEntity_save_db		(String filename);
extern Boolean		ClassEntity_load_db		(String filename);
extern EntityList	ClassEntity_all			(void);

extern Entity		Entity_create_instance		(Entity class, String title);
extern Entity		Entity_create_interface		(Entity class, String title);
extern Entity		Entity_create_class		(Entity class, String title);
extern void		Entity_free			(Entity);
extern String		Entity_title			(Entity);
extern Boolean		Entity_subclass			(Entity, Entity);  /* of, sub */
extern EntityList	Entity_get_instances		(Entity);
extern void		Entity_report_interfaceClass	(Entity);
extern String		Entity_simple_class_report	(Entity);
extern Boolean		Entity_rename			(Entity, String);
extern Boolean		Entity_load_line		(Entity *, String line);

extern Entity		string2class			(String);
extern Entity		string2entity		 	(String);
extern Entity		string2instance			(Entity superclass, String name);
extern Entity		parse_title 			(String, String*, Entity *, Boolean *);
extern Entity		Entity_lookup_translation	(Entity from_class, Entity via_class);
extern Entity		Entity_find_or_create		(String);

extern String		Entity_encodeCompactAscii	(Entity, StatusMask);
extern Boolean		Entity_decodeCompactAscii	(Entity, String *);
extern String		Entity2Report			(Entity, Boolean null_if_empty);
extern Entity		Entity_choose_interface		(Entity from, Entity to);
extern String		Entity_get_action		(Entity);

#else

extern void		ClassEntity_init		();
extern void		ClassEntity_finish		();
extern Boolean		ClassEntity_save_db		();
extern Boolean		ClassEntity_load_db		();
extern EntityList	ClassEntity_all			();

extern Entity		Entity_create_instance		();
extern Entity		Entity_create_interface		();
extern Entity		Entity_create_class		();
extern void		Entity_free			();
extern String		Entity_title			();
extern Boolean		Entity_subclass			();  /* of, sub */
extern EntityList	Entity_get_instances		();
extern void		Entity_report_interfaceClass	();
extern String		Entity_simple_class_report	();
extern Boolean		Entity_rename			();
extern Boolean		Entity_load_line		();

extern Entity		string2class			();
extern Entity		string2entity		 	();
extern Entity		string2instance			();
extern Entity		parse_title 			();
extern Entity		Entity_lookup_translation	();
extern Entity		Entity_find_or_create		();

extern String		Entity_encodeCompactAscii	();
extern Boolean		Entity_decodeCompactAscii	();
extern String		Entity2Report			();
extern Entity		Entity_choose_interface		();
extern String		Entity_get_action		();

#endif /* MONET_PROTOTYPES */

/***************************************************************************/
/***************************************************************************/
/*  $Header$ */
/***************************************************************************/
/***************************************************************************/

#endif /* _ENTITY_H_ */

