/******************************************************************/
/******************************************************************/
/* $Header$ */
/******************************************************************/
/******************************************************************/

#include "db/Entity.h"
#include "parser.h"

/******************************************************************/
/*  Globals  */
/******************************************************************/

#define HAVEACCESS(sys, id)  (id->access && (                  \
			(id->valid == a_all) ||                 \
			((id->valid & a_class ? Entity_isClass (sys): True) && \
			(id->valid & a_interface ? isInterface(sys): True))))


#define CANSAVE(sys, id) (id->cansave) && (HAVEACCESS(sys, id))


/******************************************************************/
/*  Types  */
/******************************************************************/

typedef long Integer;

static AttributeID S2A = NULL_ATTID;
static long MaskValues = 1;

/******************************************************************/
/*  Save all the user defined attributes to a file                */
/******************************************************************/

extern void ClassAttribute_save (f)
	FILE * f;
{
	AttributeID l;

	FORALL(l, S2A, next)
		if (l->standard == False)
			(void) fprintf (f, "Attribute %s %s\n", l -> string, l->help);
}

/******************************************************************/
/*  Display the attributes */
/******************************************************************/

extern void ClassAttribute_report ()
{
	AttributeID l;

	start_report ("Attributes");
	FORALL(l, S2A, next)
		if (MonetStrlen (l->string) >= 8)
			writeln_report ("%s\t- %s", l -> string, l->help);
		else
			writeln_report ("%s\t\t- %s", l -> string, l->help);
}

/******************************************************************/
/*  Encode status as compact ascii */
/******************************************************************/

static String Attribute2CompactAscii (f, code, value)
	char * f;
	char code, * value;
{
	String c;
	*f = code; ++f;
	f = MonetStringCopyToEnd (f, value);
	*f = MonitorSepCode; ++f;
	*f = '\0';
	return f;
}

extern String Entity_encodeCompactAscii (entity, mask)
	Entity entity;
	StatusMask mask;
{
	VStatus status = V_STATUS(entity);
	static char buffer [8192] = { 0 }; /* !!! */
	String f = buffer;
	AttributeID id;

	mask |= status->mask;
	FORALL(id, S2A, next)
		if ((id -> mask & mask) && HAVEACCESS(entity, id)) {
			String value;
			if (value = id->access (entity, attribute_get_shallow,
						attribute_compact_ascii, NULL))
				f = Attribute2CompactAscii (f, id->code, value);
		}
	*f = MonitorEndCode; ++f;
	*f = '\0';
	return buffer;
}

/******************************************************************/
/*  Encode status from compact ascii */
/******************************************************************/

extern StatusTrigger ChangeTrigger;
static struct status status_buffer;
static VStatus before = &status_buffer;

#define TRIG1 	if (status == NULL) return False; \
		/* status->mask = StatusNoneMask; */ \
		bcopy (status, &status_buffer, sizeof(status_buffer));
#define TRIG2	if (ChangeTrigger) { \
			ChangeTrigger (before, status, status->change_data); \
			status->mask = StatusNoneMask; \
		} \
		return True;

extern Boolean Entity_decodeCompactAscii (entity, text)
	Entity entity;
	String *text;
{
	VStatus status = V_STATUS(entity);
	String i;
	AttributeID id;

	while (SPACEORTAB(**text))
		text ++;

	if (**text == MonitorEndCode || **text == '\n' || **text == '\0')
		return True;

	TRIG1;

	for (i = *text; True; ++i) {
		char * value, code, sep;
		while (*i == MonitorSepCode)
			++i;
		code = *i;
		i++;
		value = i;
		for (; *i != NULL && *i != MonitorSepCode && *i != MonitorEndCode; ++i);
		sep = *i;
		*i = '\0';
		FORALL(id, S2A, next)
			if (id -> code == code && HAVEACCESS(entity, id)) {
				String debug;
				status->mask |= id -> mask;
				debug = id->access (entity, attribute_set, attribute_compact_ascii, value);
			}
		if (sep != MonitorSepCode)
			break;
	}
	*text = i;
	if (ChangeTrigger != NULL)
		ChangeTrigger (before, status, status->change_data);

	TRIG2;
}

/******************************************************************/
/*  Init status  */
/******************************************************************/

extern void Entity_reset_status (entity)
	Entity entity;
{
	VStatus status = V_STATUS(entity);
	StatusMask mask = status->mask;
	AttributeID id;

	if ( ! entity) {
		sys_error ("reset status, entity is null");
		return;
	}
	status->mask = 0;
	FORALL(id, S2A, next)
		if (HAVEACCESS(entity, id))
			(void) id->access (entity, attribute_unset, attribute_ascii, (String) NULL);
}


/******************************************************************/
/*  Encode status as readable ascii */
/******************************************************************/

static String Attribute2String (f, name, value)
	char * f;
	char * name, * value;
{
	String c;
	f = MonetStringCopyToEnd (f, name);
	f = MonetStringCopyToEnd (f, "= ");
	for (c=value; *c!=NULL; ++c) {
		*f = *c; ++f;
		if (*c == CONTINUATION_CHAR) {
			*f = '\n'; ++f;
			*f = '\t'; ++f;
		}
	}
	*f = '\n'; ++f;
	*f = '\0';
	return f;
}

extern String Entity2Report (entity, null_if_empty)
	Entity entity;
	Boolean null_if_empty;
{
	static char buffer [8192] = { 0 }; /* !!! */
	Attribute at;
	String name;
	AttributeID id;
	String f, place;

	f = MonetStringCopyToEnd (buffer, "Object ");
	f = MonetStringCopyToEnd (f, Entity_title (entity));
	f = MonetStringCopyToEnd (f, "\n");
	place = f;

	FORALL(id, S2A, next) {
			String value;
			if (CANSAVE(entity, id))
				if (value = id->access (entity, attribute_get_shallow, attribute_ascii, NULL))
					f = Attribute2String (f, id->string, value);
	}
	FORALL (at, entity->attributes, next) {
		char buf[32];
		if (! CANSAVE(entity, at->id))
			continue;
		name = at->id->string;
		switch (at->type) {
		case type_string:
			f = Attribute2String (f, name, at->v.string);
 			break;
		case type_boolean:
			(void) sprintf (buf, "%d", at->v.boolean);
			f = Attribute2String (f, name, buf);
 			break;
		case type_integer:
			(void) sprintf (buf, "%ld", at->v.integer);
			f = Attribute2String (f, name, buf);
 			break;
		default:
			f = Attribute2String (f, name, "");
			break;
 		}
	}
	if (place == f && null_if_empty)
		return NULL;
	*f = '\n'; ++f;
	*f = '\0'; ++f;
	return buffer;
}

/******************************************************************/
/*  Create a standard unique attibute identifier, and specify     */
/*  callback routines for getting, deleting and inserting values  */
/******************************************************************/

extern AttributeID create_attributeID (name, mask, id, standard, type, valid, access, cansave, help)
	StatusMask mask;
	AttributeType type;
	char id;
	AttributeClass valid;
	String name, help;
	AttributeRoutine access;
	Boolean standard;
	Boolean cansave;
{
	AttributeID new;

	if (String2AttributeID (name) != NULL) {
		sys_error ("Duplicate std attributeID [%s]", name);
		return NULL;
	}
	new = MonetNew (struct attributeID);
	new -> mask = mask;
	new -> string = MonetNewString(name);
	new -> type = type;
	new -> code = id;
	new -> standard = standard;
	new -> access = access;
	new -> valid = valid;
	new -> cansave = cansave;
	new -> next = S2A;
	new -> help = MonetNewString (help);
	S2A = new;
	MaskValues >> 1;
	return new;
}

/******************************************************************/
/*  Convert an attribute ID to text representation and viceversa  */
/******************************************************************/

extern AttributeID String2AttributeID (s)
	String s;
{
	AttributeID l;

	FORALL(l, S2A, next)
		if (EQS(l->string, s))
			return (l);
	return NULL_ATTID;
}

extern char * attributeID2string (id)
AttributeID id;
{
	return (id->string);
}

/******************************************************************/
/* Delete Attributes */
/******************************************************************/

extern void delete_attribute (list,a)
	AttributeID a;
	Entity list;
{
	Attribute p, previous;

	if (list==NULL)
		return;

	if (a->access != NULL) {
		(void) a->access(list, attribute_unset, attribute_ascii, NULL);
		return;
	}
	previous = list -> attributes;
	FORALL(p, list -> attributes, next) {
		if (p->id == a) {
			if (p == list -> attributes) {
				previous = list -> attributes = p -> next;
			} else {
				previous -> next = p -> next;
			}
			if (p->type == type_string)
				MonetFree (p->v.string);
			MonetFree(p);
		} else {
			previous = p;
		}
	}
}

/******************************************************************/
/* Frees a list and all its attributes */
/* Is it wise to free string values too ? */
/******************************************************************/

extern Entity free_attributeList (l)
	Entity l;
{
	Attribute a;

	FORALL (a, l -> attributes, next) {
		if (a->type == type_string) {
			free (a->v.string);   /* Is this a wise move ? */
		}
		free (a);
	}
	return NULL;
}

/******************************************************************/

/******************************************************************/

static Attribute add_attribute (l, a)
	Entity l;
	AttributeID a;
{
	Attribute new;


	if (l==NULL) {
		sys_error ("Null Attribute List");
		return NULL;
	}
	delete_attribute (l,a);  /* !!! Very inefficient */
	new = MonetNew (struct attribute);
	new -> id = a;
	new->next = l->attributes;
	l->attributes = new;
	return new;
}


extern void add_stringAttribute (l,a, s)
Entity l;
AttributeID a;
char * s;
{
	Attribute new;

	if (a->access != NULL) {
		(void) a->access (l, attribute_set, attribute_ascii, (String) s);
		return;
	}
	new = add_attribute(l,a);
	new -> type = type_string;
	new -> v.string = MonetNewString (s);
}

extern  Attribute get_attribute (l, id)
Entity l;
AttributeID id;
{
	Attribute s;

	if (l == NULL)
		return (NULL);
	if (id == NULL)
		return NULL;
	FORALL(s, l->attributes, next)
		if (s->id == id)
			return (s);

	if (l->class != NULL)
		return get_attribute (l -> class, id);
	
	return (NULL);
}

/******************************************************************/
/* Converts to string if not already string */
/******************************************************************/

extern  String get_stringAttribute (l, id)
	Entity l;
	AttributeID id;
{
	Attribute s;
	static char result [9];

	if (id->access != NULL)
		return ( (String) id->access (l, attribute_get, attribute_ascii, NULL));
	if (id->type != type_string)
		return NULL;
	if ((s = get_attribute (l, id)) == NULL)
		return NULL;
	switch (s->type) {
	case type_integer:
		(void) sprintf (result, "%ld", s->v.integer);
		return result;
	case type_string:
		return s->v.string;
	case type_boolean:
		if (s->v.boolean)
			strcpy (result, "True");
		else
			strcpy (result, "False");
		return result;
	default:
		return NULL;
	}
	/* Not reached */
}

extern Boolean present_attribute (l, id)
Entity l;
AttributeID id;
{
	if (id == NULL) {
		sys_error ("NULL id");
		return False;
	}
	if (id->access != NULL)
		return (True);   /*  !!! Dubious */

	return (get_attribute(l,id) != NULL);
}

/******************************************************************/
/*  Initialise and finish */
/******************************************************************/

extern void init_attributes ()
{
	S2A = NULL;
}

extern void finish_attributes ()
{
	AttributeID s;

	FORALL (s, S2A, next) {
		MonetFree (s->string);
		MonetFree (s->help);
		MonetFree (s);
	}
	S2A = NULL;
}

/******************************************************************/
/******************************************************************/
/* $Header$ */
/******************************************************************/
/******************************************************************/

