/***************************************************************/
/***************************************************************/
/*  $Header$  */
/***************************************************************/
/*  Since all attributes are dealt with in roughly the same way*/
/*  the basic access operation on them is defined by a macro.  */
/*  Now this is pretty difficult to write, let alone debug, so */
/*  a better way should be found.                              */
/***************************************************************/

#include "db/Entity.h"
#include "parser.h"
#include "at_translation.h"

extern void set_shapeOf_entity ();
extern String save_instance_type ();
extern String save_shape_attribute ();

StandardAttributes Std_Attribute;

/***************************************************************/
/*  Get next sequence of non space chars in a string  */
/***************************************************************/

static Boolean get_string (result, buf)
	String * buf, result;
{
	String place, start;
	place = *buf;
	for (;*place != NULL && (*place == ' ' || !isprint (*place)); place++);
	start = place;
	for (;*place != NULL && *place != ' ' && isprint (*place); place++, result++) {
		*result = *place;
	}
	if (place == start)
		return False;
	*buf = place;
	*result = NULL;
	return True;
}

/*****************************************************************/
/*****************************************************************/

extern String access_component (sys, op, format, list)
	Entity sys;
	AttributeOperation op;
	AttributeFormat format;
	String list;
{
	Entity s;
	static char buffer[1024] = { 0 }; /* !!! */
	String place = buffer;
	int cc;
	Boolean first;

	if (op == attribute_unset) {
		free_components (sys);
		return NULL;
	} else if (op == attribute_set) {
		Entity subcomp;
		char sub[1024]; /* !!! */
		String next;
		next = list;
		while (next) {
			if ( ! get_string (sub, &next))
				break;
			if (subcomp = Entity_find_or_create (sub)) 
				make_subcomponent (sys, subcomp);
			for (;*next != NULL && *next != CONTINUATION_CHAR; next++);
			for (;*next != NULL && *next == CONTINUATION_CHAR; next++);
		}
		return list;
	}
	if (isEmpty_entityList (sys->subcomponents))
		return NULL;
	first = True;
	FOR_ENTITYLIST (s, sys->subcomponents) {
		if (first)
			first = False;
		else
			place = MonetStringCopyToEnd (place, " $ ");
		place = MonetStringCopyToEnd (place, Entity_title (s));
	}
	return buffer;
}

/*****************************************************************/
/* Set interfaces with a string, no difference between these     */
/* These are called when parsing the source files */
/*****************************************************************/

extern String access_interface (interface, op, format, interface_to)
	Entity interface;
	AttributeOperation op;
	AttributeFormat format;
	String interface_to;
{
	Entity draw_interface;
	char interface1[1024], interface2[1024];
	int cc;

	if (op == attribute_unset) {
		unmake_interface (interface, FROM(interface));
		unmake_interface (interface, TO(interface));
		set_neighbour (V_STATUS(interface));
		return NULL;
	} else if (op == attribute_set) {
		cc = sscanf (interface_to, "%s %s", interface1, interface2);
		if (cc == 2) {
			unmake_interface (interface, FROM(interface));
			unmake_interface (interface, TO(interface));
		}
		if (cc >= 1) {
			if ((draw_interface = Entity_find_or_create (interface1)) == NULL) {
				return NULL;
			} else {
				make_interface (interface, draw_interface);
			}
			if (cc == 2 && (draw_interface = Entity_find_or_create (interface2)))
				make_interface (interface, draw_interface);
			set_neighbour (V_STATUS(interface));
			return interface_to;
		}
	}
	if (FROM(interface) || TO(interface)) {
		static char buffer[1024] = { 0 }; /* !!! */
		String place = buffer;
		if (FROM(interface))
			place = MonetStringCopyToEnd (buffer, Entity_title (FROM(interface)));
		/* -- Place now points to last position in buffer */
		if (TO(interface)) {
			place = MonetStringCopyToEnd (place, " ");
			(void) MonetStringCopyToEnd (place, Entity_title (TO(interface)));
		}
		return buffer;
	}
	return NULL;
}

extern String get_ping_instruction (o)
	Entity o;
{
	String ping;

	if (ping = access_ping (o, attribute_get, attribute_ascii, NULL))
		/*  Replace any & with the title of the entity  */
		return MonetReplaceInString (ping, "&", TITLE(o));

	return NULL;
}

/***************************************************************/
/*  generic function for accessing attributes  */
/***************************************************************/

static String char2String (c)
	char c;
{
	static char buf[2] = " ";
	*buf = c;
	return buf;
}

#define SETSTRING(ACCESS_MACRO, ID, value) if (ACCESS_MACRO(o)) MonetFree (ACCESS_MACRO(o)); \
	ACCESS_MACRO(o) = MonetNewString (value); MASK(o) |= ID;

#define SETORD(ACCESS_MACRO, ID, value)    if (ACCESS_MACRO(o) != value) { ACCESS_MACRO(o) = value; MASK(o) |= ID; }

#define ACCESSSTRING(routine_name, ID, ACCESS_MACRO) \
		ACCESS(routine_name, ID, ACCESS_MACRO, \
			,,,, \
			String, \
			SETSTRING, NULL)

#define ACCESSORD(routine_name, ID, ACCESS_MACRO, TOSTRING, FROMSTRING, TYPE, UNSET) \
		ACCESS(routine_name, ID, ACCESS_MACRO,  \
			TOSTRING, FROMSTRING,            \
			TOSTRING, FROMSTRING, \
			TYPE, \
			SETORD, UNSET)

#define ACCESSENM(routine_name, ID, ACCESS_MACRO, TOSTRING, FROMSTRING, TYPE, UNSET) \
	 	ACCESS(routine_name, ID, ACCESS_MACRO,    \
			TOSTRING, FROMSTRING, \
			char2String, *,                     \
			TYPE,                                \
			SETORD, UNSET)

#define ACCESS(routine_name, ID, ACCESS_MACRO, TOSTRING, FROMSTRING, TOCOMPACT, FROMCOMPACT, TYPE, SET, UNSET) \
extern String routine_name (o, op, format, data) \
	Entity o; \
	AttributeOperation op; \
	AttributeFormat format; \
	String data; \
{ \
	TYPE value;                          \
                                              \
	if ( ! o) {                           \
		return NULL;                   \
	} else if (op == attribute_unset) {     \
		SET(ACCESS_MACRO, ID, UNSET);         \
	} else if (op == attribute_set) {           \
		if (format == attribute_compact_ascii) \
			if (data)                                    \
				value = (TYPE) (FROMCOMPACT(data)); \
			else                    \
				value = UNSET; \
		else \
			value = FROMSTRING(data);            \
		SET(ACCESS_MACRO, ID, value);             \
	} else if (op == attribute_get_shallow) {     \
	} else if (op == attribute_get) {                 \
		if (ACCESS_MACRO(o) == UNSET)              \
			if (o->class)                       \
				o = o->class;                \
		/* If cannot inherit from class, inherit from parent !!!      \
		if (ACCESS_MACRO(o) == UNSET)                  \
			if (o->parent)                          \
				o = o->parent; */                \
	} else {                                                  \
		sys_error ("Unknown attribute access operation [%d]", (int) op); \
	}                                                    \
	if (ACCESS_MACRO(o) == UNSET)		           \
		return NULL;                                \
	if (format == attribute_compact_ascii)               \
		return TOCOMPACT(ACCESS_MACRO(o));                     \
	return TOSTRING((TYPE)ACCESS_MACRO(o));                     \
}

/***************************************************************/
/*  For accessing specific attributes  */
/***************************************************************/

/* ACCESSSTRING(access_action,	); */
ACCESSSTRING(access_title,	StatusTitleMask, TITLE);
ACCESSSTRING(access_ping,	StatusPingMask,PING);
ACCESSSTRING(access_contact,	StatusContactMask,CONTACT);

ACCESSENM(access_reachability,	StatusReachMask,STATUS,		reachstatus2string,	string2reachstatus,	ReachStatus,	 status_unknown);
ACCESSENM(access_future,	StatusFutureMask,FUTURE,		future2string,		string2future,		FutureStatus,	 FUTURE_DEFAULT);
ACCESSENM(access_behaviour,	StatusBehaviourMask,BEHAVIOUR,		behaviour2string,	string2behaviour,	Behaviour, behaviour_normal);

ACCESSORD(access_connections,	StatusConnectionsMask,CONNECTIONS,	cardinal2string,	string2cardinal,	Cardinal,  NUM_CONNECTIONS_UNKNOWN);
ACCESSORD(access_load_in,	StatusLoadInMask,LOADIN,	load2string,		string2load,		Load,	  UNKNOWN_LOAD);
ACCESSORD(access_load_out,	StatusLoadOutMask,LOADOUT,	load2string,		string2load,		Load,	 UNKNOWN_LOAD);

ACCESSORD(access_capacity,	StatusCapacityMask,CAPACITY,	capacity2string,	string2capacity,	Capacity,  UNKNOWN_CAPACITY);
ACCESSORD(access_octets_in,	StatusOctInMask, OCTIN,		capacity2string,	string2capacity,	Capacity,	 UNKNOWN_CAPACITY);
ACCESSORD(access_octets_out,	StatusOctOutMask,OCTOUT,	capacity2string,	string2capacity,	Capacity,	 UNKNOWN_CAPACITY);
ACCESSORD(access_pkts_in,	StatusPktInMask,PKTIN,		capacity2string,	string2capacity,	Capacity,	 UNKNOWN_CAPACITY);
ACCESSORD(access_pkts_out,	StatusPktOutMask,PKTOUT,	capacity2string,	string2capacity,	Capacity,	 UNKNOWN_CAPACITY);

ACCESSSTRING(access_action,	StatusActionMask,ENTACTION);

/***************************************************************/
/*  Get action for entity  */
/***************************************************************/

extern String Entity_get_action (entity)
	Entity entity;
{
	static char buffer[1024];
	char *request, *c;
	String action;
	if (action = access_action (entity, attribute_get, attribute_ascii, NULL)) {	
		/*  Replace any & with the title of the entity  */
		for (c = action, request=buffer; *c!=NULL; c++) {
			if (*c == '&') {
				/*** If a '&' is found in the template then substitute the systems title/address ***/
				request = MonetStringCopyToEnd (request, TITLE (entity));
			} else {
				if (*c == '\\')
					c++;
				/*** Copy the character in the template into the request string ***/
				*request = *c;
				request ++;
			}
		}
		*request = NULL;
		return buffer;
	}
	return NULL;
}

/***************************************************************/
/*  Init and finish  */
/***************************************************************/

extern void init_class ()
{
#define ASTR(name, mask, id, op, help) create_attributeID (name, mask, id, True, type_string, a_all, op, True, help);
#define AINT(name, mask, id, op, help) create_attributeID (name, mask, id, True, type_integer, a_all, op, True, help);
#define ACHR(name, mask, id, op, help) create_attributeID (name, mask, id, True, type_char, a_all, op, True, help);

	/* Does nothing with these */
	Std_Attribute.format = create_attributeID (
		"format", StatusFormatMask, ' ',	True, type_string,	a_class, set_titleFormat, True,
		"Title must match this pattern");
	Std_Attribute.shape = create_attributeID (
		"shape", StatusShapeMask, ' ',	True, type_string,	a_class, set_shapeOf_entity, True,
		"Shape used when drawn on map");
	Std_Attribute.resizable = create_attributeID (
		"resizable", StatusResizeMask, ' ',	True, type_flag,	a_class, NULL, True,
		"Can be resized" );
	Std_Attribute.anonymous	= create_attributeID (
		"anonymous", StatusAnonMask, ' ',	True, type_flag,	a_class, set_anonymous_name, True,
		"Name will be generated randomly");
	Std_Attribute.unbendingInterface = create_attributeID (
		"unbending", StatusUnbendingMask, ' ',	True, type_flag,	a_class, set_unbendingInterface, True,
  		"Interface will be straight");
	Std_Attribute.underlying =	create_attributeID (
		"underlying",  StatusUnderlyingMask, ' ',	True, type_flag,a_class, set_underlying, True, 
		"This can be an underlying component of another entity");
	Std_Attribute.instances	= create_attributeID (
		"instances", StatusInstancesMask,  ' ',	True, type_string,	a_class, set_instances, True,
		"Classes instances are maps, systems, itnerfaces or none");
	(void) create_attributeID (
		"translation", StatusTranslationMask, ' ', True, type_string, a_class, set_translation, True,
		"Rules for translating one class of map into another");

	Std_Attribute.title =
		create_attributeID ("title", StatusTitleMask, 't', True, type_string, a_all, access_title, False,
		"The title of an entity (domain name)");
	Std_Attribute.ping = ASTR("ping", StatusPingMask, 'p', access_ping,
		"Instruction for monitoring entities");
	(void) ASTR("contact",		StatusContactMask, 'c', access_contact,
		"Who to contact in case of problems");
	(void) ACHR("reach",		StatusReachMask, 'r', access_reachability,
		"Reachability of an entity");
	(void) ACHR("future",		StatusFutureMask, 'f', access_future,
		"Future of an entity, temporary, certain, planned, ...");
	(void) ACHR("behaviour",	StatusBehaviourMask, 'b', access_behaviour, 
		"The behaviour of an entity, normal, chaotic, ...");
	(void) AINT("capacity",		StatusCapacityMask, 'C', access_capacity,
		"Capacity of entity, usually speed");
	(void) AINT("connections",	StatusConnectionsMask, '#', access_connections,
		"Number of connections to an entity.");
	(void) AINT("loadin",		StatusLoadInMask, 'i', access_load_in,
		"Percentage traffic in over capacity for a very recent period");
	(void) AINT("loadout",		StatusLoadOutMask, 'o', access_load_out,
		"Percentage traffic out over capacity for a very recent period");
	(void) AINT("octin",		StatusOctInMask, 'I', access_octets_in,
		"Total number of bytes in");
	(void) AINT("octout",		StatusOctOutMask, 'O', access_octets_out,
		"Total number of bytes out");
	(void) AINT("pktin",		StatusPktInMask, 'P', access_pkts_in,
		"Total number of packets in");
	(void) AINT("pktout",		StatusPktOutMask, 'Q',  access_pkts_out,
		"Total number of packets out");
	(void) ASTR("action",		StatusActionMask, 'a', access_action,
		"A <monetsh> command, executed when entity is clicked");
  	(void) create_attributeID ("interface",	StatusInterfaceMask, 'i', True, type_string,
		a_interface,
		access_interface,
		True, 
		"Two entities interfaced together");
	(void) ASTR("component",	StatusComponentMask, 'z', access_component,
		"List of component entities of an entity");

}

extern void finish_class ()
{
}

/***************************************************************/
/***************************************************************/
/*  $Header$  */
/***************************************************************/
/***************************************************************/

