/******************************************************************/
/******************************************************************/
/* $Header: //gaia/group.local/computer/monet/Draw/RCS/ps.c,v 1.1 1991/03/26 10:52:24 a38 Exp a38 $ */
/******************************************************************/
/******************************************************************/

#include "general/General.h"
#include "general/Draw.h"

/******************************************************************/
/*  Globals  */
/******************************************************************/

DrawLibrary DrawPSLib;

static FILE * DumpFile = NULL;
static float ReverseMask = 0;

/******************************************************************/
/*  Send message to dump file  */
/******************************************************************/

static void OUT (a,b,c,d,e,f,g,h,i)
	char * a;
{
	fprintf (DumpFile, a,b,c,d,e,f,g,h,i);
	fprintf (DumpFile, "\n");
}

/******************************************************************/
/*  Copy the prologue file to the output                          */
/******************************************************************/

static Boolean out_prologue (name)
{
	char * filename;
	char data [1024]; /* !!! */
	FILE * fid = NULL;

	filename = (char *) MonetFilename (Config_maps_postscript_prolog);

	OUT ("%% --- Page Format --- ");
	OUT ("/monetformat (%s) def", name);
	OUT ("%% ------------------- ");

	fid = fopen (filename, "r");
	if (fid == NULL) {
		user_warning ("Cannot open postscript prologue file [%s]", filename);
		return False;
	}

	while (fgets(data, sizeof(data), fid) != NULL) {
		fputs (data, DumpFile);
	}

	fclose (fid);
	return True;
}


/******************************************************************/
/*  Write out the grey shade/colour  */
/******************************************************************/

static void OUTSHADE (shade)
	Shade shade;
{
	float value;

	switch (shade) {
	case shade_0:
		value = 1.0;
		break;
	case shade_1:
		value = 0.75;
		break;
	case shade_2:
		value = 0.5;
		break;
	case shade_3:
		value = 0.25;
		break;
	case shade_4:
		value = 0.0;
		break;
	default:
		value = 0.5;
		sys_error ("Unknown shade [%d]", shade);
		break;
	}
	OUT ("%f monetsetshade", ReverseMask - value);
}



/******************************************************************/
/* Before drawing anything on a particular map, this routine must */
/* be called */
/******************************************************************/

extern Boolean start_draw_postscript (prologue_name, title, filename, xl, xr, yl, yr)
	String title, prologue_name, filename;
	Position xl,xr, yl,yr;
{
	PROC("start_draw_postscript");

	DumpFile = fopen (filename, "w");
	if (DumpFile == NULL) {
		user_error ("Could not open file [%s]", filename);
		return False;
	} else {
		user_inform ("Dumping 'postscript' format map to file [%s]", filename);
	}
	ReverseMask = 0.0;

	OUT ("%%! %s", Config_software_title);
	OUT ("%%%%BoundingBox: %d %d %d %d", xl, yr, xr, yl);
	if (prologue_name == NULL) {
		OUT ("%% This picture has no headers or footers");
		OUT ("%% and is for inclusion in other documents");
		prologue_name = "A_.PS";
		
	} else {
		OUT ("%% This picture has headers and footers");
	}
	OUT ("%%%%Creator: %s", Config_maps_author);
	OUT ("%%%%Title: %s", title);
	OUT ("%%%%CreationDate: %s", pretty_time_stamp ());
	OUT ("%%%%Pages: 1");
	OUT ("%%%%DocumentFonts: Times-Roman Helvetica");
	OUT ("%%%%EndComments");

	OUT ("%% Parameters for the prologues");
	OUT ("/monettimestamp      (%s) def", pretty_time_stamp ());
	OUT ("/monettitle          (%s) def", title);
	OUT ("/monetauthor         (%s) def", Config_maps_author);
	OUT ("/monetcontact        (%s) def", Config_maps_contact);
	OUT ("/monetsource         (%s) def", Config_maps_source);
	OUT ("/monetsoftwaretitle  (%s) def", Config_software_title);
	OUT ("/monetsoftwaresource (%s) def", Config_software_source);
	out_prologue (prologue_name);

	OUT ("%d %d %d %d monetscalemap", xl,yl, xr,yr);
	OUT ("%%Page: 0 1");

	Draw_start_draw (DrawPSLib, 1.0, 1.0);
	return True;
}

static void end_draw ()
{
	OUT("showpage");
	OUT ("");
	fclose (DumpFile);
}


/* !!! Should be draw_fraction line */
static void 
draw_percentage_line (shade, numer, denom, xl, yl, xr, yr)
	Position        xl, yl, xr, yr;
	Shade shade;
	Cardinal numer, denom;
{
	Position x,y;  /*  The intermediate point */

	OUTSHADE (shade);
	if (numer == 0)
		OUT (" %d %d %d %d monetline",  xl,yl, xr,yr);
	else {
		if (xl>xr)
			x = (xl-xr)*numer/denom + xr;
		else
			x = (xr-xl)*numer/denom + xl;
		if (yl>yr)
			y = (yl-yr)*numer/denom + yr;
		else
			y = (yr-yl)*numer/denom + yl;
		OUT (" %d %d %d %d %d %d monetpercentageline",  xl,yl, xr,yr, x,y);
	}
}

static void 
draw_line (shade,  xl, yl, xr, yr)
	Position        xl, yl, xr, yr;
	Shade shade;
{
	OUTSHADE (shade);
	OUT (" %d %d %d %d monetline",  xl,yl, xr,yr);
}

static void 
fill_elipse (shade, x, y, width, height)
	Shade shade;
	Position        x, y;
	Dimension       width, height;
{
	OUTSHADE (shade);
	OUT (" %d %d %d %d monetfillelipse",  x,y, width,height);
}

static void 
draw_elipse (shade, x, y, width, height)
	Shade shade;
	Position        x, y;
	Dimension       width, height;
{
	OUTSHADE (shade);
	OUT (" %d %d %u %u monetdrawelipse",  x,y, width,height);
}


static void 
draw_rectangle (shade, x, y, width, height)
	Shade shade;
	Position        x, y;
	Dimension       width, height;
{
	OUTSHADE (shade);
	OUT (" %d %d %u %u monetdrawrectangle",  x,y, (int) width,(int)height);
}

static void 
draw_roundedRectangle (shade, x, y, width, height)
	Shade shade;
	Position        x, y;
	Dimension       width, height;
{
	OUTSHADE (shade);
	OUT (" %d %d %u %u monetdrawrectangle",  x,y, (int) width,(int)height);
}

static void 
fill_roundedRectangle (shade, x, y, width, height)
	Shade shade;
	Position        x, y;
	Dimension       width, height;
{
	OUTSHADE (shade);
	OUT (" %d %d %u %u monetfillrectangle",  x,y, (int) width,(int)height);
}

static void 
fill_rectangle (shade, x, y, width, height)
	Shade shade;
	Position        x, y;
	Dimension       width, height;
{
	OUTSHADE (shade);
	OUT (" %d %d %d %d monetfillrectangle",  x,y, width,height);
}

static void 
draw_box (shade, x, y, width, height, xspace, yspace)
	Shade shade;
	Position        x, y;
	Dimension       width, height, xspace, yspace;
{
	fill_rectangle (shade, x,y, width, height);
 	if ((width > 2*xspace) && (height > 2*yspace))
		fill_rectangle (shade_background, x+xspace,y+yspace, (Dimension)(width-2*xspace),  (Dimension) (height-2*yspace));
}

static void 
draw_string (shade, x, y, string)
	Shade shade;
	Position        x, y;
	String string;
{
	OUTSHADE (shade);
	OUT (" (%s) %d %d monetstring",  string, x,y);
}

static void  draw_bitmap (id, x,y)
	BitmapId id;
	Position x, y;
{
}

static void  register_bitmap	(id, w,h, bits)
	BitmapId id;
	Dimension w, h;
	char * bits;
{
}

static void set_scale (x,y)
	float x,y;
{
	/*  Scale text  */
	OUT (" %f %f monettextscale", x,y);
}

static void set_reverse ()
{
	if (ReverseMask > 0.5)
		ReverseMask = 0.0;
	else
		ReverseMask = 1.0;
}

/*****************************************************************/
/* Init and finish drawing of entitys */
/*****************************************************************/

extern void 
init_draw_postscript ()
{
	PROC("init_draw_postscript");
	NEWDRAWLIB(DrawPSLib);
}

extern void 
finish_draw_postscript()
{
	MonetFree (DrawPSLib);
	DrawPSLib = NULL;
}


/******************************************************************/
/******************************************************************/
/* $Header: //gaia/group.local/computer/monet/Draw/RCS/ps.c,v 1.1 1991/03/26 10:52:24 a38 Exp a38 $ */
/******************************************************************/
/******************************************************************/



