/*****************************************************************************/
/*****************************************************************************/
/* $Header: //gaia/group.local/computer/monet/Xmonet/RCS/main.c,v 1.1 1991/03/26 12:00:14 a38 Exp a38 $ */
/*****************************************************************************/
/* This module provides support for a simple (extensible) script programming */
/* language.  It provides basic functionality, loops, conditionals, vars,    */
/* pipes, macros and a set of standard functions.  Other modules can extend  */
/* on the set of standard functions using the C interface.                   */
/* The main bias is towards short scripts for text processing, especially    */
/* processing the output for normal operating system programs, so that use   */
/* sh, awk etc can be avoided when they aren't really necessary and impose   */
/* an unwanted processor overhead.                                           */
/* Input can come from a file, stdin or from a string.                       */
/* !!! This stuff is very similar to the code for config and they could be   */
/* !!! combined.                                                             */
/*****************************************************************************/

#include "script.h"
#include "general/Help.h"
#include "general/Channel.h"
#include "io/io.h"
#include <signal.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/ioctl.h>

/***************************************************************************/
static MonetFinishRoutine Finish;

struct script_macro {
	String name;
	Cardinal args;
	ScriptCallback callback;
	String value;
	String help, details;
	Cardinal depth;
	ScriptMacro next;
};

static ScriptMacro ScriptMacroStack;
static Cardinal ScriptMacroDepth = 0;

static String execute PR((String, String *, String, String));
static String ScriptMacro_call PR((String label, String name, String *arg));

static UserMessageRoutine User;
static WriteReportRoutine Write;

/*****************************************************************************/
/*  Standard print routine  */
/*****************************************************************************/

static String CurrentPipeCommand = NULL;

static void user_message_pipe (label, a,b,c,d,e,f,g,h,i)
     char * a, * label;
{
	if (CurrentPipeCommand) {
		(void) execute (CurrentPipeCommand, NULL, NULL, NULL);
	} else {
		User (label, a,b,c,d,e,f,g,h,i);
	}
}

/*************************************************************************/
/*  Write to the report  */
/*************************************************************************/

static void write_report_pipe (a,b,c,d,e,f,g,h,i)
	char * a;
{
	if (CurrentPipeCommand) {
		(void) execute (CurrentPipeCommand, NULL, NULL, NULL);
	} else {
		Write (a,b,c,d,e,f,g,h,i);
	}
}

/***************************************************************************/
/*  Syntax analysis  */
/***************************************************************************/

#define FREE_MALLOCED_ARGS \
	while (malloced_argc>0) \
		{--malloced_argc; MonetFree (malloced_arg[malloced_argc]); }
static String execute (command, prev_arg, pipe, pipe_input)
	String command;
	char *prev_arg[];
	String pipe, pipe_input;
{
	String label, name, value, statement;
	int cc, c;
	char * cp;
	char *arg[MAXSCRIPTARG], *malloced_arg [MAXSCRIPTARG];
	Cardinal argc = 0, malloced_argc = 0;
	LexSymbol lex;
	String result = NULL;
	Boolean okay = True;
	String left= NULL;
	Boolean left_is_seq = False;

	if (MonetStrlen (command) == 0)
		return NULL;
	for (argc = 0; argc<MAXSCRIPTARG; argc++)
		arg[argc] = NULL;

	cp = command;
	/*  -- Ignore whitespace  */
	while (*cp == ' ' || *cp == '\t')
		cp++;

	value = Script_next_symbol_loose (&cp, &lex);
	while (okay) {
                switch (lex) {
		case lex_statement:
			left = value;
			left_is_seq = True;
			value = Script_next_symbol_loose (&cp, &lex);
			break;
		case lex_name:
		case lex_op2:
			FREE_MALLOCED_ARGS;
			argc = 0;
			name = value;
			left_is_seq = False;
			left = name;
			while (True) {
				value = Script_next_symbol_loose (&cp, &lex);
				if (lex == lex_sep || lex == lex_end || lex == lex_pipe ||
					lex == lex_and || lex == lex_or ||
					lex == lex_pipe_in || lex == lex_pipe_out)
					break;
        		        switch (lex) {
				case lex_label:
				case lex_string:
				case lex_name:
				case lex_op2:
					arg[argc] = value;
/*					debug_message ("arg[%u] = [%s]", argc, arg[argc]); */
					++ argc;
					break;
				case lex_statement:
					arg[argc] = malloced_arg[malloced_argc] =
						MonetNewString (
							execute (value, NULL, NULL, NULL));
					++ argc;
					++ malloced_argc;
					break;
				case lex_parameter:
					arg[argc] = prev_arg[atoi(value)];
/* 					debug_message ("arg[%u] = [%s]", argc, arg[argc]); */
					++ argc;
					break;
				case lex_pipe_parameter:
					arg[argc] = pipe_input;
/*					debug_message ("arg[%u] = [%s]", argc, arg[argc]); */
					++ argc;
					break;
				default:
					user_error ("Syntax error  [%s]", value);
					value = Script_next_symbol_loose (&cp, &lex);
					break;
				}
				arg[argc] = NULL;
			}
			break;
		case lex_label:
			label = value;
			value = Script_next_symbol_loose (&cp, &lex);
			break;
		case lex_pipe_in:
		case lex_pipe_out:
		case lex_and:
		case lex_or:
		case lex_pipe:
			if ( ! left)
				user_error ("Found [%s], must have a left hand side", value);
			else {
				value = Script_next_symbol_loose (&cp, &lex);
			}
		case lex_sep:
		case lex_end:
			if (lex == lex_pipe) {
				sys_error ("Pipe found");/*
				User = set_user_inform_routine (user_message_pipe);
				Write = set_write_report_routine (write_report_pipe);

				pipe = CurrentPipeCommand;
				CurrentPipeCommand = value;*/
			}
			if (left) {
				if (left_is_seq) {
					ScriptMacro_deeper ();
					result = execute (left, arg, NULL, NULL);
					CurrentPipeCommand = pipe;
					ScriptMacro_shallower ();
				} else {
					result = ScriptMacro_call (label, left, arg);
				}
				left = NULL;
			}
			if (lex == lex_end)
				okay = False;
			else
				value = Script_next_symbol_loose (&cp, &lex);
			break;
		default:
			user_error ("Syntax error  [%s]", value);
			value = Script_next_symbol_loose (&cp, &lex);
			break;
		}
		if (lex == lex_and && !result)
			okay = False;
		if (lex == lex_or && result)
			okay = False;
	}
	FREE_MALLOCED_ARGS;
	return result;
}

/***************************************************************************/
/*  Managing macros  */
/***************************************************************************/

static String ScriptMacro_call (label, name, arg)
	String name, *arg, label;
{
	Boolean found;
	ScriptMacro macro;
	String result = NULL;

	if (MonetStrlen (name) == 0)
		return NULL;

	/*  -- Seach list of known commands  */

	found = False;

	FORALL (macro, ScriptMacroStack, next) {
		if (EQS(name, macro->name)) {
       			if (macro->callback == NULL) {
/*				debug_message ("Expanding [%s]", macro->name); */
				if (macro->value)
					result = execute (macro->value, arg, NULL, NULL);
			} else {
/*				debug_message ("Calling [%s]", macro->name); */
				result = macro->callback (arg);
			}
			found = True;
		} else if (EQS(name, "exit")) {
			found = True;
		}
	}
	if ( ! found)
		user_inform ("Don't recognise command [%s]", name);
	return result;
}

extern Script_write (pipe, arg)
	String pipe;
	String *arg;
{
	if (! pipe) {
		while (*arg) {
			write_report (*arg);
			++arg;
		}
		writeln_report ("\n");
	} else {
		execute (pipe, NULL, NULL, *arg);
	}
}

/***************************************************************************/
/*  Creating macros  */
/***************************************************************************/

extern ScriptMacro ScriptMacro_create (name, value, callback, help, details)
	String name, value, help, details;
	ScriptCallback callback;
{
	ScriptMacro macro;

	macro = MonetNew (struct script_macro);
	macro -> name = MonetNewString (name);
	macro -> callback = callback;
	macro -> value = MonetNewString (value);
	macro -> help = MonetNewString (help);
	macro -> details = MonetNewString (details);
	macro -> depth = ScriptMacroDepth;
	LIST_ADD (macro, ScriptMacroStack, next);
	return macro;
}

extern void ScriptMacro_deeper ()
{
	ScriptMacroDepth ++;
}

extern void ScriptMacro_shallower ()
{
	ScriptMacro macro;

	FORALL (macro, ScriptMacroStack, next)
		if (macro->depth < ScriptMacroDepth)
			break;
		else {
			MonetFree (macro);
			MonetFree (macro->help);
			MonetFree (macro->details);
			MonetFree (macro->value);
			MonetFree (macro->name);
		}

	ScriptMacroStack = macro;
	ScriptMacroDepth --;
}

/***************************************************************************/
/***************************************************************************/
/*  Standard functions  */
/***************************************************************************/
/***************************************************************************/

/***************************************************************************/
/*  User defined macros  */
/***************************************************************************/

SCRIPTCALLBACK(do_define)
	String name, help, value;

	name = arg[0];
	value = arg[1];
	if (MonetStrlen (name) == 0 || MonetStrlen (value) == 0) {
		user_error ("def <name> <value> [<help>]");
	} else {
		if (arg[2])
			help = arg[2];
		else
			help = "user defined";
		debug_message ("define [%s] as [%s], help= [%s]", name, value, help);
		ScriptMacro_create (name, value, NULL, help, NULL);
	}
ENDSC

/***************************************************************************/
/*  Finish off */
/***************************************************************************/

SCRIPTCALLBACK(do_quit)
	if (Finish)
		Finish ();
ENDSC

/***************************************************************************/
/* Help messages */
/***************************************************************************/

SCRIPTCALLBACK(do_help)
	String what = arg[0];
	ScriptMacro macro;
	Cardinal depth;

	writeln_report (
	"For details of all defined functions, use 'syntax', or for a specific");
	writeln_report (
	"command, 'syntax <function>'");
		writeln_report (
	"Functions can be grouped by enclosing them in ( and ).  Arithmetic is");
		writeln_report (
	"prefix and can only be done inside (...).  Inside (...) strings must be");
		writeln_report (
	"surrounded by quotes.  A sequence of functions can be seperated by,");
		writeln_report (
	"\t&&\tIf left succesful then do right");
		writeln_report (
	"\t||\tIf left unsuccesful then do right");
		writeln_report (
	"\t;\tDo left then right");
/*		writeln_report (
	"\t|\tPipe output from left into right");
		writeln_report (
	"$0,$1 .. $9 are replaced by actual arguments");
		writeln_report (
	"%%0,%%1 .. %%9 are replaced by piped input");
		writeln_report (
	"All instruction sequence can be preceded by a label");
		writeln_report (
	"Use 'exit <label>' to leave sequence");
		writeln_report (
	"\teg :abc {loop; select * ; exit :abc}");
 !!! */
ENDSC

SCRIPTCALLBACK(do_syntax)
	String what = arg[0];
	ScriptMacro macro;
	Cardinal depth;

	if (what && * what) {
		FORALL(macro, ScriptMacroStack, next)
			if (EQS(what, macro->name)) {
				writeln_report (" %s\t - %s", macro->name, macro->help);
				if (MonetStrlen(macro->details))
	  				writeln_report (" \t\t%s", macro->details);
				break;
			}
		if (macro == NULL)
			writeln_report ("Sorry, no help on [%s]", what);
	} else {
		writeln_report ("Syntax of Commands");
		depth = ScriptMacroDepth;
		FORALL(macro, ScriptMacroStack, next) {
			if (macro->depth < depth) {
				writeln_report ("\t-----------------");
				depth = macro->depth;
			}
			if (MonetStrlen (macro->name) < 8)
				writeln_report (" %s\t\t\t- %s", macro->name, macro->help);
			else if (MonetStrlen (macro->name) < 16)
				writeln_report (" %s\t\t- %s", macro->name, macro->help);
   			else
				writeln_report (" %s\t- %s", macro->name, macro->help);
			if (MonetStrlen(macro->details))
  				writeln_report (" \t\t%s", macro->details);
		}
	}
ENDSC

/*****************************************************************************/
/*  Read and execute commands  */
/*****************************************************************************/
                               
extern String Script_execute (command)
	String command;
{
	char *arg[MAXSCRIPTARG];
	Cardinal argc;
	String result;

	for (argc = 0; argc<MAXSCRIPTARG; argc++)
		arg[argc] = NULL;

/*	User  = set_user_inform_routine (user_message_pipe);
	Write = set_write_report_routine (write_report_pipe);
*/
	debug_message ("Execute command [%s]", command);
	result = execute (command, arg, NULL, NULL);
/*	if (MonetStrlen (result))
		writeln_report ("Returned [%s]", result);
*/
/*	(void) set_user_inform_routine (User);
	(void) set_write_report_routine (Write);
*/
	return result;
}

/*****************************************************************************/
/*  Buffer input from stdin  */
/*****************************************************************************/

extern Boolean Script_execute_fd (fid, interactive)
	int fid;
	Boolean interactive;
{
	char command[MAXSCRIPTCOMMANDLEN]; /* !!! */
	Cardinal place = 0;
	char c;
	int cc;

	while (1) {
		cc = read (fid, &c, 1);
		if (cc == 0) {
			user_inform ("End of file");
			return False;
		} else if (cc == -1) {
			if (errno == EWOULDBLOCK || errno == EAGAIN) {
				command[place] = NULL;
				debug_message ("Fd %d would block, buffer [%s]", fid, command);
			 } else if (errno == ENOTCONN) {
				sys_error ("Fd is not connected");
			} else {
				sys_error ("Bad recv: errno [%d]", errno);
			}
			break;
		} else 	if (c == '\n' || place == MAXSCRIPTCOMMANDLEN - 1) {
			command[place] = NULL;
			(void) Script_execute (command);
			if (interactive) {
				printf  ("%s> ", Config_program_name);
				fflush (stdout);
			}
			place = 0;
			break;
		}
		command[place] = c;
		place ++;
	}
	if (interactive)
		fflush (stdout);  /* !!! Questionable */
	return True;
}

/***************************************************************************/
/*  If statement  */
/***************************************************************************/

SCRIPTCALLBACK(do_if)
	if (execute (arg[0], NULL, NULL, NULL))
		return execute (arg[1], NULL, NULL, NULL);
ENDSC

SCRIPTCALLBACK(do_loop)
	while (execute (arg[0], NULL, NULL, NULL));
ENDSC

/*****************************************************************************/
/*  Init finish  */
/*****************************************************************************/

extern void Script_init (finish)
	MonetFinishRoutine finish;
{
	if (finish == NULL)
		sys_error ("Finish routine is NULL");

	ScriptMacro_deeper ();
	ScriptMacro_create ( "quit",	NULL,	do_quit,	 	"quit", "");
	ScriptMacro_create ( "q",	NULL,	do_quit,	 	"quit", "");
	ScriptMacro_create ( "exit",	NULL,	do_quit,	 	"quit", "");
	ScriptMacro_create ( "help",	NULL,	do_help, 		"help information" , "");
 	ScriptMacro_create ( "syntax",	NULL,	do_syntax, 		"command syntax" , "");
	ScriptMacro_create ( "def",	NULL,	do_define, 		"Define a macro" , "");
	ScriptMacro_create ( "if",	NULL,	do_if,		 	"if {condition} {do}" , "");
	ScriptMacro_create ( "loop",	NULL,	do_loop,		 "loop {sequence}" , "");

	Script_init_base ();

	Finish = finish;
}

extern void Script_finish ()
{
	Script_finish_base ();
	ScriptMacro_shallower ();
}

/*****************************************************************************/
/*****************************************************************************/
/* $Header: //gaia/group.local/computer/monet/Xmonet/RCS/main.c,v 1.1 1991/03/26 12:00:14 a38 Exp a38 $ */
/*****************************************************************************/
/*****************************************************************************/

