/*****************************************************************************/
/*****************************************************************************/
/* $Header: //gaia/group.local/computer/monet/Xmonet/RCS/main.c,v 1.1 1991/03/26 12:00:14 a38 Exp a38 $ */
/*****************************************************************************/
/*  Lexical stuff for monet script  */
/*  Could maybe better use lex                               */
/*****************************************************************************/

#include "script.h"

/*****************************************************************************/
/*  Lexical Scanner  */
/*  returns the next item  */
/*****************************************************************************/

static String get_until (source, chr)
	String *source;
	char chr;
{
	String result, cp;

	for (result = cp = (*source)+1; *cp && *cp != chr; ++cp);

	if (*cp == '\0') {
		user_error ("End of command found where [%c] expected", chr);
		*source = cp;
	} else {
		*cp = NULL;
		*source = cp+1;
	}

	return result;
}

static String get_nested (source, first, chr)
	String *source;
	char chr;
{
	String result, cp;
	Cardinal nest = 0;

	for (result = cp = (*source)+1; True; ++cp) {
		if (*cp == first) {
			++nest;
		} else if (*cp == chr) {
			if (nest == 0) {
				*cp = NULL;
				*source = cp+1;
				break;
			}
			-- nest;
		} else if (*cp == NULL) {
			user_error ("End of command found where [%c] expected", chr);
			*source = cp;
			break;
		}
	}
	return result;
}


static Boolean in (c, set)
	char c, * set;
{
	while (set && *set)
		if (c == *set)
			return True;
		else
			++set;
	return False;
}

extern String Script_next_symbol_tight (source, lex)
	LexSymbol *lex;
	String *source;
{
	return Script_next_symbol_loose (source, lex);
}

extern String Script_next_symbol_loose (source, lex)
	LexSymbol *lex;
	String *source;
{
	String result = *source;
	String cp = *source, c;
	char tmp;

	if ( ! cp) {
		*lex = lex_end;
		return result;
	}
	/*  Ignore comments */
	while (*cp == '#')
		for (; *cp && *cp != '\n'; ++cp);

	switch (*cp) {
	case '\0':
		*lex = lex_end;
		break;
	case '\'':
		result = get_until (&cp, '\'');
		*lex = lex_string;
		break;
	case '\\':
		++cp;
		result = cp;
		*lex = lex_string;
		++cp;
		break;
	case '\"':
		result = get_until (&cp,  '"');
		*lex = lex_string;
		break;
	case '(':
		result = get_nested (&cp,  '(', ')');
		*lex = lex_statement;
		break;
	case '$':
		result = cp+1;
		*lex = lex_parameter;
		cp += 2;
		break;
	case '>':
		*lex = lex_pipe_out;
		++cp;
	case '<':
		*lex = lex_pipe_in;
		++cp;
		break;
	case '|':
		if (*cp == '|') {
			*lex = lex_or;
			++cp;
		} else
			*lex = lex_pipe;
		break;
	case '+':
	case '/':
	case '%':
	case '-':
	case '*':
		result = cp;
		++cp;
		*lex = lex_op2;
		break;
	case '&':
		if (*cp == '&') {
			*lex = lex_and;
			++cp;
			break;
		} else {
			user_error ("Unexpected symbol [%s]", cp);
			*lex = lex_sep;
		}
		break;
	case '?':
		result = cp;
		*lex = lex_parameter;
		cp += 1;
	case '\n':
	case ';':
		/*  -- ignore blanks and command serarators */
		while (*cp == ' ' || *cp == '\t' || *cp == '\n' || *cp == ';')
			cp++;
		*lex = lex_sep;
		break;
	case ':':
		++cp;
		result = cp;
		while (*cp && isalnum (*cp))
			++cp;
		*lex = lex_label;
		break;
	default:
		if (isalnum (*cp) || *cp == '_') {
			result = cp;
			while (*cp && isgraph (*cp) && ! in (*cp, "&|;"))
				++cp;
			*lex = lex_name;
		} else {
			user_error ("Unexpected symbol [%s]", cp);
			++cp;
			*lex = lex_sep;
		}
		break;
	}

	if (*cp == ' ' || *cp == '\t') {
		/* --  Set last character in result to NULL */
		*cp = '\0';
		/*  -- Ignore whitespace  */
		do {
			++cp;
		} while (*cp == ' ' || *cp == '\t');
	} else if (*cp != '\0') {
		/* -- Shift everything up to make room for NULL char */
		for (c = cp, tmp = *c; tmp; ++c) {
			char tmp2;
			tmp2 = *(c+1);
			*(c+1) = tmp;
			tmp = tmp2;
		}
		*(c+1) = '\0';
		/* --  Set last character in result to NULL */
		*cp = '\0';
		++cp;
	}

	*source = cp;
/*	debug_message ("token %u, value [%s]", *lex, result); */
	return result;
}



/*****************************************************************************/
/*****************************************************************************/
/* $Header: //gaia/group.local/computer/monet/Xmonet/RCS/main.c,v 1.1 1991/03/26 12:00:14 a38 Exp a38 $ */
/*****************************************************************************/
/*****************************************************************************/
