
/***************************************************************************/
/***************************************************************************/
/*  $Header: //gaia/group.local/computer/monet/FileServer/RCS/server.c,v 1.1 1991/05/06 08:17:59 a38 Exp a38 $  */
/***************************************************************************/
/*  This sits and waits with select for new network events.  It is entirely */
/*  syncronous, using no timing interupts.  */
/*  Currently it multiplexes new connections, but I would like to extend it */
/*  to listen for all the pings and requests etc.  */
/***************************************************************************/

#include "general/General.h"
#include "general/Channel.h"
#include "io/io.h"
#include <errno.h>
#include <sys/file.h>
#include <sys/ioctl.h>

/***************************************************************************/
/*  Private definitions  */
/***************************************************************************/

struct input_fid {
	int fid;
	FidHandler handler;
	caddr_t data;
	FidType type;
	InputFid next;
};

static InputFid InputFidList = NULL;

static fd_set Readbits, Writebits;
static int Maxfid = -1;

#define SETMAXFID(fd) if(fd>Maxfid) Maxfid = fd;

/***************************************************************************/
/*  Initialises a server, supply port number and callbacks for new         */
/*  connections, disconnections and messages,  returns an identifier       */
/***************************************************************************/

extern Boolean start_server ()
{
	FD_ZERO(&Readbits);
	FD_ZERO(&Writebits);
	return True;
}

/***************************************************************************/
/*  Closes a server */
/***************************************************************************/

extern void close_server ()
{
}

/***************************************************************************/
/*  Handle list of inputs waiting for select  */
/***************************************************************************/

extern InputFid create_input_fid (fid, type, handler, data)
	int fid;
	FidHandler handler;
	FidType type;
	caddr_t data;
{
	InputFid new;

	new = MonetNew (struct input_fid);
	new -> fid = fid;
	new -> handler = handler;
	new -> data = data;
	new -> next = InputFidList;
	new -> type = type;
	InputFidList = new;
	SETMAXFID(fid);

	if (fid < 0) {
		/* bad or psuedo fid */
	} else if (type == fid_read)
		FD_SET(fid, &Readbits);
	else if (type == fid_write)
		FD_SET(fid, &Writebits);
	else
		sys_error ("Unknown Fid Type [%d]", type);

	return new;
}

extern void free_input_fid (fid)
	int fid;
{
	InputFid old = NULL, i;

	FORALL(i, InputFidList, next)
		if (i->fid == fid)
			old = i;
	if (old == NULL) {
		sys_error ("input fid not found %d", fid);
		return;
	}
	if (fid < 0) {
		/* bad or psuedo fid */
	} else if (old->type == fid_read)
		FD_CLR(fid, &Readbits);
	else if (old->type == fid_write)
		FD_CLR(fid, &Writebits);
	else
		sys_error ("Unknown Fid Type [%d]", old->type);

	LIST_DEL (InputFid, old, InputFidList, next);
	close (fid);
	MonetFree (old);
}


/***************************************************************************/
/*  Wait for Messages from the client, send requests and return answers.   */
/***************************************************************************/

extern void wait_for_messages (timenow, period, before, after)
	long period;
	long timenow;
	BeforeSelectHandler before;
	AfterSelectHandler after;
{
	fd_set readbits, writebits, other;
	struct timeval timer;
	long start;
	int count, tablesize, numbits;

	PROC ("wait_for_messages");

	tablesize = getdtablesize();
	timerclear(&timer);
	timer.tv_sec = period;


	while (True) {
		InputFid i;
		bcopy ((char *)(&Readbits),  (char *)(&readbits),  sizeof(Readbits));
		bcopy ((char *)(&Writebits), (char *)(&writebits), sizeof(Writebits));
		FD_ZERO(&other);

		numbits = tablesize;
		start = time (0);
		debug_message ("Starting select for %d seconds", timer.tv_sec);

		if (before)
			before (&numbits, &readbits, &writebits, &other, &timer);

		do {
			count = select(numbits, &readbits, &writebits, &other, &timer);
		} while (count < 0 && errno == EINTR);

		if (after)
			after (count, readbits, writebits, other);

		if (count < 0) {
			perror("select");
		} else if (count == 0) {
			 /* Timer ran out */
				break;
		} else if (count > 0){
			for (i = InputFidList; i != NULL;) {
				if (i->fid < 0) {
					/*  Special connections, ie non-socket  */
					i = i->next;
				} else {
					/*  A connection may be removed so we need  */
					/*  to remember where the 'next one is.     */
					InputFid next;
					next = i->next;
					if (FD_ISSET (i->fid, &readbits)) {
						i->handler (i->fid, i->data);
					} else if (FD_ISSET (i->fid, &writebits)) {
						i->handler (i->fid, i->data);
					}
					i = next;
				}
			}
		}
		start = time (0) - start;
		if (start >= period)
			break;
		timerclear(&timer);
		timer.tv_sec = start;
	}
}

/***************************************************************************/
/***************************************************************************/
/*  $Header: //gaia/group.local/computer/monet/Monetd/RCS/server.c,v 1.2 1991/03/22 16:09:27 a38 Exp a38 $  */
/***************************************************************************/
/***************************************************************************/

