/***************************************************************/
/***************************************************************/
/* $Header: //gaia/group.local/computer/monet/lib/map/RCS/attach.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $  */
/***************************************************************/
/*  Seems to do everything I want it to, might be nice if */
/*  it was more inteligent and tried to avoid other boxes */
/*  Would be nice f it drew an absolute 45 degree line rather */
/*  than another angle. */
/*  Could do partial adjustments better */
/*  A totally straight line is better than two at right angles */
/*  !!! Bug:  Does not properly clip lines to edges of boxes */
/******************************************************************/

#include "Drawing.h"
#include "Layout.h"

/***************************************************************/
/***************************************************************/

/* !!! These should be elsewhere, in configure */
#define Config_x_attachMargin 2
#define Config_y_attachMargin 2

/***************************************************************/
/* Return the difference between big and small or 0 if big <small */
/***************************************************************/

#define diffCard(big,small) (big<small?0:big-small)
 
static void nearest_point_of_2_lines (x1, x2, w1,w2, xr1, xr2)
   Position        *xr1, *xr2, x1, x2;
   Dimension       w1, w2;
{
   /* If not overlapping */ 
   if (x1+w1 <= x2) {
      *xr1 = x1+w1;
      *xr2 = x2;
   } else if (x1 >= x2+w2) {
      *xr1 = x1;
      *xr2 = x2+w2;
   /*  Else overlapping, use mid point of ovrlapping region */
   } else if (x1 <= x2) {
      if (x1+w1 <= x2+w2)
        *xr1 = *xr2 = x2 + (w1 - x2 + x1)/2;
      else
        *xr1 = *xr2 = x1 + w1/2;
   } else if (x1 > x2) {
      if (x1+w1 > x2+w2)
        *xr1 = *xr2 = x1 + (w2 - x1 + x2)/2;
      else
        *xr1 = *xr2 = x1 + w1/2;
  }
}

/************************************************************************/
/* Choose attchment points */
/************************************************************************/

/* Chooses the 'best' line between two drawings */
extern void choose_attachment(from, to, xl, yl, xr, yr, straight, xstraight, length)
	Drawing        from, to;
	Position       *xl, *yl, *xr, *yr;
	Boolean        *straight, *xstraight;
	Dimension      *length;
{
   Position        xx, yy, x1, y1, x2, y2;
   Dimension       w1, w2, h1, h2;

	if (Drawing_mapOf (from) != Drawing_mapOf (to)) {
		/*  They are not on the same map, 'to' is on the parent map */
	}

	x1 = X(from);
	y1 = Y(from);
	w1 = WIDTH(from);
	h1 = HEIGHT(from);

	x2 = X(to);
	y2 = Y(to);
	w2 = WIDTH(to);
	h2 = HEIGHT(to);

   /* Adjust so that straight lines never hit very end of box */
   /* Pretty ad hoc */
   if (w2 < Config_x_attachMargin * 2) {
      x2 = x2 + w2 / 2;
      w2 = 0;
   } else {
      x2 += Config_x_attachMargin;
      w2 -= Config_x_attachMargin * 2;
   }
   if (h2 < Config_y_attachMargin * 2) {
      y2 = y2 + h2 / 2;
      h2 = 0;
   } else {
      y2 += Config_y_attachMargin;
      h2 -= Config_y_attachMargin * 2;
   }

   nearest_point_of_2_lines (x1,x2, w1,w2, xl, xr);
   nearest_point_of_2_lines (y1,y2, h1,h2, yl, yr);

   /*   Need to trim the lines by the margin+1 */
   if (*xr == *xl) {
      *xstraight = True;
      *length = abs (yl - yr);
      *straight = True;
   } else if (*yr == *yl) {
      *length = abs (xl - xr);
      *xstraight = True;
      *straight = True;
   } else {
      *length = 0;  /* !!! Undefined */
      *straight = False;
   }
}



/* Finds a 'good' place to position drawing 'mid', somewhere between
   'sys1' and 'sys2'.  If the current position is 'better' it will
   return the current position.

   if the new position is on a straight line choose that, else
   if the previous place had straight lines between at least one of the
   systems do not use the new value.
*/

extern void choose_midpoint_attachment (mid, sys1,sys2, x, y)
	Drawing mid, sys1, sys2;
	Position *x, *y;
{
	Position xl,yl, xr,yr;
	Boolean     straight,     straight1,     straight2, isx1, isx2, isx;
	Dimension len, len1, len2;

	choose_attachment (mid, sys1, &xl,&yl, &xr,&yr, &straight1, &isx1, &len1);
	choose_attachment (mid, sys2, &xl,&yl, &xr,&yr, &straight2, &isx2, &len2);
	choose_attachment (sys1, sys2, &xl,&yl, &xr,&yr, &straight, &isx,  &len);
	if ((!straight || (straight && len > len1 + len2)) && straight1 && straight2) {
		user_inform ("Keep position [%s]", Drawing_title(mid));
		*x = X(mid);
		*y = Y(mid);
	} else if (!straight && straight1) {
		user_inform ("Alter position sys2 [%s]", Drawing_title(mid));
	/* Should straighten connection to sys2 */
	if (isx1) { /* Alter the y */
		*y = Y(sys2) + HEIGHT(sys2)/2;
		*x = X(mid);
	} else { /* Alter the x */
		*x = X(sys2) + WIDTH(sys2)/2;
		*y = Y(mid);
		}
	} else if (!straight && straight2) {
		user_inform ("Alter position sys1 [%s]", Drawing_title(mid));
		/* Should straighten connection to sys1 */
		if (isx2) {
			*y = Y(sys1) + HEIGHT(sys1)/2;
			*x = X(mid);
		} else {
			*x = X(sys1) + WIDTH(sys1)/2;
			*y = Y(mid);
		}
	} else {
		/* user_inform ("New position [%s]", Drawing_title(mid)); !!! */
		*x = (xl + xr)/2;
		*y = (yl + yr)/2;
	}
}

extern void choose_midpoint_position (sys1, sys2, x,y)
	Drawing sys1, sys2;
	Position *x, *y;
{
	Position  xl,yl, xr,yr;
	Boolean   straight, isx;
	Dimension len;

	choose_attachment (sys1, sys2, &xl,&yl, &xr,&yr, &straight, &isx,  &len);
	/* user_inform ("New position [%s]", Drawing_title(mid)); !!! */
	*x = (xl + xr)/2;
	*y = (yl + yr)/2;
}

/***************************************************************/
/***************************************************************/
/* $Header: //gaia/group.local/computer/monet/lib/map/RCS/attach.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $  */
/***************************************************************/
/***************************************************************/

