
/*****************************************************************************/
/*****************************************************************************/
/* $Header: //gaia/group.local/computer/monet/lib/map/RCS/map.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $  */
/* $Log: map.c,v $
 * Revision 1.1  1991/06/11  11:26:10  a38
 * Initial revision
 *  */
/*****************************************************************************/
/*  For maps.  This part tries to be totally independent of X windows, it    */
/*  defines only the purely structural aspects.  See map.c for X stuff  */
/*****************************************************************************/

#include "Drawing.h"
#include "Interface.h"
#include "Script.h"

/*****************************************************************************/
/* Globals */
/*****************************************************************************/

static Map MapList = NULL;

/***********************************************************/
/*  Map changed or resized */
/***********************************************************/

static MapChangedCallback Change_callback;
static Boolean CanChange = True;

extern void Map_set_resized (map)
	Map map;
{
	if (CanChange) {
		map -> saved = False;
		if (Change_callback)
			Change_callback (map);
	}

}

extern void Map_set_changed (map)
	Map map;
{
	if (CanChange) {
		map -> saved = False;
	}
}

extern void	Map_disable_set_changed		()
{
	CanChange = False;
	temporary_noDrawing ();
}

extern void	Map_enable_set_changed		()
{
	CanChange = True;
	allow_drawing ();
}


/******************************************************************/
/*  Set position for grid */
/******************************************************************/

extern void Map_set_grid (map, x,y)
	Map map;
	Cardinal x, y;
{
	map->grid_x = x;
	map->grid_y = y;
}

/******************************************************************/
/*  Adjust position for grid */
/******************************************************************/

extern void Map_adjust_for_grid (map, x,y, newx,newy)
	Map map;
	Position x,y, *newx, *newy;
{
	*newx = (x >> map->grid_x) << map->grid_x;
	*newy = (y >> map->grid_y) << map->grid_y;
}

/*****************************************************************************/
/*  Return the map widget  */
/*****************************************************************************/

extern Widget Map_widget (map)
	Map map;
{
	while (map)
		if (map->mapWidget)
			return map->mapWidget;
		else if (map->parent)
			map = Drawing_mapOf (map->parent);
		else
			return NULL;
	return NULL;
}

/*****************************************************************************/
/*  Return the root map  */
/*****************************************************************************/

extern Map Map_root (map)
	Map map;
{
	while (map)
		if (map->parent)
			map = Drawing_mapOf (map->parent);
		else
			return map;
	return NULL;
}

/*****************************************************************************/
/*  Background file with graphics commands  */
/*****************************************************************************/

extern void Map_set_background_file	(map, filename)
	Map map;
	String filename;
{
	if (map->background_file)
		MonetFree (map->background_file);
	map->background_file = MonetNewString (filename);
}

extern String Map_get_background_file		(map)
	Map map;
{
	return map->background_file;
}

extern void Map_set_background_colour	(map, colourname)
	Map map;
	String colourname;
{
	if (map->background_colour)
		MonetFree (map->background_colour);
	map->background_colour = MonetNewString (colourname);
}

extern String Map_get_background_colour		(map)
	Map map;
{
	return map->background_colour;
}

/*****************************************************************************/
/*  Traverse list of maps  */
/*****************************************************************************/

extern Map Map_first ()
{
	Map map;
	map = MapList;
	return map;
}

extern Map Map_next (current)
	Map current;
{
	/*  assume current != NULL  */
	return current->next;
}

/*****************************************************************************/
/*  Return False if there are maps still to be saved.                        */
/*****************************************************************************/

extern Boolean AreAllMapsSaved()
{
	Map          m;
	FORALLMAPS (m)
		if (m->saved == False)
			return False;
	return True;
}
/*****************************************************************************/
/* Redraw all the maps  */
/*****************************************************************************/

extern void ClassMap_redraw_all()
{
	Map         m;
	FORALLMAPS (m)
		Map_request_redraw (m);
}

/*****************************************************************************/
/*  Class maps changed  */
/*****************************************************************************/

extern void ClassMap_changed()
{
	Map         m;
	FORALLMAPS (m)
		if (Map_classesOnly(m))
			Map_request_redraw (m);
}

/*****************************************************************************/
/*  Position elements on a map drawn as a list  */
/*****************************************************************************/

#define CLASS_X_SPACING 16
#define CLASS_Y_SPACING 5

static Position list_class(map, class, x,y)
	Entity          class;
	Position        x, y;
	Map map;
{
	Entity          c;
	Drawing draw;

	if (class == NULL) {
		sys_error ("list_class class is NULL");
		return y;
	}
	draw = Map_insert_drawing_class (map, class, 0, 0);
	if (!draw) {	
		sys_error ("Could not get drawing for class [%s]", Entity_title(class));
		return y;
	}
	LISTBOX(draw).x = x;
	LISTBOX(draw).y = y;
	Map_adjust_corners (Drawing_mapOf(draw), draw, True);
	Drawing_reset_list_label (draw);
	y += LISTBOX(draw).height + CLASS_Y_SPACING;
	x += CLASS_X_SPACING;
	FOR_ENTITYLIST(c, Entity_subclasses(class))
		y = list_class(map, c, x, y);
	return (y);
}

extern void setup_list_map (map)
	Map map;
{
	if (Map_classesOnly(map)){
		list_class (map, TopClass, 20, 10);
	} else {
		Drawing o;
		Position y = 30;
		FOR_DRAWINGLIST(o, Map_elements (map)) {
			LISTBOX(o).x = 20;
			LISTBOX(o).y = y;
			Drawing_reset_list_label (o);
			y += LISTBOX(o).height + CLASS_Y_SPACING;
		}
	}

}

/*****************************************************************************/

/*****************************************************************************/

extern void Map_set_drawn_as (map, draw)
	Map map;
	DrawnAs draw;
{
	if (map->drawnas == draw)
		return;

	map->drawnas = draw;

	switch (map->drawnas) {
	case drawnas_list:
		setup_list_map (map);
		break;
	default:
	case drawnas_map:
		break;
	}
	Map_set_changed (map);
	Map_request_redraw (map);
}

/*****************************************************************************/
/*  Draw a new drawing on a map that has just been placed there              */
/*  ie for mapareas drawn as lists, redraw the entire map                    */
/*****************************************************************************/

static void Drawing_draw_new (s)
	Drawing s;
{
	Map map = Drawing_mapOf(s);
	if ( ! isDrawingAllowed ())
		return;

	if (Map_classesOnly(map)) {
		Map_request_redraw (map);
	} else {
		switch (map->drawnas) {
		case drawnas_list:
			setup_list_map (map);
			Map_request_redraw (map);
			break;
		case drawnas_map:
			Drawing_redraw(s);
			break;
		}
	}
}

/*****************************************************************************/
/*  UnDraw an old drawing on a map that has just been removed from there     */
/*****************************************************************************/

static void Drawing_draw_old (s)
	Drawing s;
{
	Map map = Drawing_mapOf(s);
	if (isDrawingAllowed ()) {
		if (Map_classesOnly(map)) {
			Map_request_redraw (map);
		} else {
			setup_list_map (map);
			switch (map->drawnas) {
			case drawnas_list:
				Map_request_redraw (map);
				break;
			case drawnas_map:
				Drawing_undraw(s);
				break;
			}
		}
	}
}

/*****************************************************************************/
/*  Set a default domain for a map.  This way labels are kept short, because */
/*  the default is not displayed.  */
/*****************************************************************************/

extern void Map_set_domain (map, domain)
	Map map;
	String domain;
{
	if (NES(domain, map->default_domain)) {
		Drawing d;
		if (map->default_domain)
			MonetFree (map->default_domain);
		map->default_domain = MonetNewString (domain);
		map->default_domain_len = MonetStrlen (map->default_domain);
		FOR_DRAWINGLIST (d, Map_elements (map)) {
			/* -- inefficient, need to sort out all label stuff */
			reset_label (d);
		}
	}
}

extern String Map_get_domain (map)
	Map map;
{
	return map -> default_domain;
}

/*****************************************************************************/
/* Tests wether or not a particular class of drawing can be placed in a       */
/* particular map. */
/*****************************************************************************/

extern Boolean Map_can_insert_drawing (map, class)
	Entity class;
	Map map;
{
	if (map->entity == NULL) {
		return False;
	} else if (Entity_subclass (Std_Type.annotation, class)) {
	    /* Annotation can go anywhere */
	} else if (class == Std_Type.button) {
	    /* Buttons can go anywhere */
	} else if (class == Std_Type.interface) {
	    /* !!! This shouldn't be allowed */
	} else if (Entity_subclass (
			Entity_classOf(
				Entity_classOf(map->entity)),
			class) == False) {
		return False;
	}
	return True;
}

/************************************************************************/
/*  Backup coordinates  */
/************************************************************************/

extern void Map_backup (map)
	Map map;
{
	Drawing o;
	FOR_DRAWINGLIST (o, Map_elements (map)) {
		DRAWING(o)->backup.x = X(o);
		DRAWING(o)->backup.y = Y(o);
		DRAWING(o)->backup.width = WIDTH(o);
		DRAWING(o)->backup.height = HEIGHT(o);
	}
}

extern void Map_restore (map)
	Map map;
{
	Drawing o;
	FOR_DRAWINGLIST (o, Map_elements (map)) {
		X(o) = DRAWING(o)->backup.x;
		Y(o) = DRAWING(o)->backup.y;
		WIDTH(o) = DRAWING(o)->backup.width;
		HEIGHT(o) = DRAWING(o)->backup.height;
	}
	Map_request_redraw (map);
}

/*****************************************************************************/
/* This is the move system instruction, to be called whenever the position of
 * a system is changed */
/*****************************************************************************/

extern Boolean Map_insert_drawing (map, new, x, y, w, h)
	Drawing new;
	Map map;
	Position x,y;
	Dimension w,h;
{
	Map m;

	/*****************************************************************************/
	/*  Check parameters  */
	/*****************************************************************************/

	if (new == NULL) {
		sys_error ("Can't insert NULL drawing into a map");
		return False;
	}
	if (map == NULL) {
		sys_error ("NULL map");
		return False;
	}
	for (m = map; m && m->parent; m = m->parent->map)
		if (new == m->parent) {
			user_error ("Cannot insert maps into their children");
			return False;
		}
	if (Drawing_classOf(map) == NULL) {
		sys_error ("Cannot have a map with a NULL class");
		return False;
	}

	/* This is a little silly, ought to have a special routine for classes !!! */
	if (Map_classesOnly(map)) {
		if (! Entity_isClass (ENTITY(new))) {
			sys_error ("Cannot put an instance [%s] into class map",
				Drawing_title (new));
			return False;
		}
	} else if (Drawing_isClass (new)) {
		sys_error ("Cannot put a class [%s] into a map other than class_map",
				Drawing_title (new));
		return False;
	}

	if (Map_can_insert_drawing (map, Drawing_classOf (new)) == False) {
		String map_title = MonetNewString (Map_title (map));
		user_error ("Cannot insert [%s] in map [%s]", Drawing_title (new), map_title);
		MonetFree (map_title);
		return False;
	}

	/*****************************************************************************/
	/* Deal with map part first */
	/*****************************************************************************/

	if (Drawing_mapOf(new) == map) {
		Drawing_undraw(new);
	} else {
		if (Drawing_mapOf(new)) {
      			Map_extract_drawing (Drawing_mapOf(new), new);
		}
		Drawing_mapOf(new) = map;
		Drawing_prepare(new);
		insert_drawingList(Map_elements (map), new);
	}

	/*****************************************************************************/
	/*  Now work out where the drawing goes on the map,  x,y,w,h                 */
	/*****************************************************************************/

	if (x<=0 || y<= 0) {
		if (Drawing_isInterface (new))
			(void) get_newInterfacePosition  (map, new, &x, &y);
		else {
			if (x<=0)
				x = map->unplaced_x;
			if (y<=0)
				y = map->unplaced_y += HEIGHT(new) + 10;
		}
	}
	Drawing_dimension (new, x,y,w,h);

	/*****************************************************************************/
	/*  Now reflect the changes on the map  */
	/*****************************************************************************/

	Drawing_draw_new (new);
 	Map_set_changed (map);

	/*****************************************************************************/
	/* !!!	(void) position_interfaces (new);  Should work */
	/*	straighten_acute_interfaces (new);   !!! */
	/*****************************************************************************/

	return True;
}

/*****************************************************************************/
/*  Extract a drawing from a map */
/*****************************************************************************/

extern void Map_extract_drawing (map, new)
	Map map;
	Drawing          new;
{
	if (new == NULL)
		return;

	if (map == Drawing_mapOf(new)) {
		Map_adjust_corners (map, new, False);
		Drawing_draw_old (new);
		extract_drawingList(Map_elements (map), new);
	 	Map_set_changed (map);
		Drawing_mapOf(new) = NULL;
	} else {
		sys_error ("map does not correspond to argument");
	}
}

/*****************************************************************************/
/*  Free/Extract all drawings of a map */
/*****************************************************************************/

extern void Map_extract_all_drawings (map)
	Map map;
{
	Drawing d;

	temporary_noDrawing ();
	while (d=first_drawingList (map->elements)) {
		Drawing_free (d);
	}
	allow_drawing ();
 	Map_set_changed (map);
	Map_request_redraw (map);
}

/***************************************************************/
/*  Title of a map  */
/***************************************************************/

extern String Map_title (map)
	Map map;
{
	if (map)
		return Entity_title (map->entity);
	return NULL;
}

/***********************************************************/
/* Check the placement of a drawing and adjust the corners */
/* of the map if necessary.  */
/***********************************************************/

extern void Map_adjust_corners (map, drawing, adding)
	Map map;
	Drawing drawing;
	Boolean adding;
{
	Position x,y;
	Dimension w,h;
	Boolean change;

	if (map != Drawing_mapOf (drawing)) {
		sys_error ("map doesn't match");
		return;
	}
	change = False;
	switch (map->drawnas) {
	case drawnas_list:
		x = LISTBOX(drawing).x;
		y = LISTBOX(drawing).y;
		w = LISTBOX(drawing).width;
		h = LISTBOX(drawing).height;
		break;
	case drawnas_map:
		x = X(drawing);
		y = Y(drawing);
		w = WIDTH(drawing);
		h = HEIGHT(drawing);
		break;
	}
	/*  Resize the map area  */
	if (adding) {
		if (x < map->xl)	{ change = True; map->xl = x; }
		if (x+w > map->xr)	{ change = True; map->xr = x+w; }
		if (y < map->yl)	{ change = True; map->yl = y; }
		if (y+h > map->yr)	{ change = True; map->yr = y+h; }
	} else {
		/* -- if we are removing this extreme node */
		/* -- then invalidate map size  */
		if (x <= map->xl)   { change = True; map->xl = 0; }
		if (x+w >= map->xr) { change = True; map->xr = 0; }
		if (y <= map->yl)   { change = True; map->yl = 0; }
		if (y+h >= map->yr) { change = True; map->yr = 0; }
	}
	if (change)
		Map_set_resized (map);
}

/***********************************************************/
/*  Find the extreme corners of a map, ie the left and right */
/*  most positions of the drawings on it.  Returns false if  */
/*  the map is empty, else true.   */
/***********************************************************/

extern Boolean Map_corners (map, xl,yl, xr,yr, force_check)
	Map map;
	Position *xl, *yl, *xr, *yr;
	Boolean force_check;
{
	Drawing s;
	Boolean change = False;

	*xl = map->xl;
	*xr = map->xr;
	*yl = map->yl;
	*yr = map->yr;

	if (!force_check && ARESET_DIMENSIONS(map))
		return True;

	FOR_DRAWINGLIST (s, Map_elements (map)) {
		Position x,y;
		Dimension w,h;

		switch (map->drawnas) {
		case drawnas_list:
			x = LISTBOX(s).x;
			y = LISTBOX(s).y;
			w = LISTBOX(s).width;
			h = LISTBOX(s).height;
			break;
		case drawnas_map:
			x = X(s);
			y = Y(s);
			w = WIDTH(s);
			h = HEIGHT(s);
			break;
		}
		if (x < *xl)	{change = True; *xl = x; }
		if (x+w > *xr)	{change = True; *xr = x+w; }
		if (y < *yl)	{change = True; *yl = y; }
		if (y+h > *yr)	{change = True; *yr = y+h; }
	}
	if (! change && ! ARESET_DIMENSIONS(map)) {
		if (map->parent) {
			*xr = WIDTH(map->parent);
			*yr = HEIGHT(map->parent);
		} else {
			*xr = 1;
			*yr = 1;
		}
		*xl = 1;
		*yl = 1;
	} else {
		map->xl = *xl;
		map->xr = *xr;
		map->yl = *yl;
		map->yr = *yr;
		if (change)
			Map_set_resized (map);
	}
	return True;
}

extern void Map_discard (map)
	Map map;
{
	Drawing drawing;
	FOR_DRAWINGLIST(drawing, Map_elements (map))
		if (drawing->submap)
			Map_free (drawing->submap);
	Map_extract_all_drawings (map);
	map->entity = NULL;
}

/* !!! Should free the entity if no one else is referencing it */
/* !!! Should put a reference counter into entity structure */

extern void Map_free (map)
	Map map;
{
	Map_discard (map);
	LIST_DEL(Map, map, MapList, next);
	reset_drawingList (map->elements);
	Drawing_freeList (map->elements);
	MonetFree (map->background_file);
	MonetFree (map->background_colour);
	MonetFree (map);
}

extern Map Map_create ()
{
	Map new;

	if (new = MonetNew(struct map)) {
		LIST_ADD(new, MapList, next);
		new -> marked = False;
		new -> scale = NULL;
		new -> drawnas = drawnas_map;
		new -> entity = NULL;
		new -> unplaced_x = 20;
		new -> unplaced_y = 20;
		new -> xl = 0;
		new -> xr = 0;
		new -> yl = 0;
		new -> yr = 0;
		new -> show_load = True;
		new -> class_map = False;
		new -> elements = Drawing_createList ();
		new -> saved = True;  /* !!! Hack, because reset in position_new */
		new -> grid_x = 3;
		new -> grid_y = 3;
		new -> default_domain = NULL;
		new -> default_domain_len = 0;
		new -> background_file = NULL;
		new -> background_colour = NULL;
		new -> connectors = connector_default;
		new -> parent = NULL;
		/*  !!!  These fields are available to systems using X  */
		/*  !!!  Not referenced in this module.  */
		new -> mapWidget = NULL;
		new -> clip_region = NULL;
		new -> x_scale = 1.0;
		new -> y_scale = 1.0;
		return new;
	} else {
		sys_error ("Couldn't malloc space for new map area");
	}
	return False;
}

/***********************************************************/
/*  Initialise / finish off */
/***********************************************************/

extern void ClassMap_init (change)
	MapChangedCallback change;
{
	Change_callback = change;
	ClassEntity_init ();
	/*  -- Load default database  */
	ClassEntity_load_db (Config_class_edb);

	MapList = NULL;
	map_server_init ();
	MapScript_init (Config_finish_handler);

	ClassDrawing_init ();
	init_vstatus(); /* temp */
	init_mark();
	init_layout ();

	init_drawMaps ();
}

extern void ClassMap_finish ()
{
	Map m;

	finish_drawMaps();
	finish_layout ();
	finish_mark ();
	finish_vstatus ();

	MapScript_finish ();
	FORALLMAPS (m)
		Map_free (m);
	map_server_finish ();
	ClassEntity_finish ();
}

/*****************************************************************************/
/*****************************************************************************/
/* $Header: //gaia/group.local/computer/monet/lib/map/RCS/map.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $  */
/*****************************************************************************/
/*****************************************************************************/

