/******************************************************************************/
/******************************************************************************/
/*  $Header: //gaia/group.local/computer/monet/lib/map/RCS/mark.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $ */
/******************************************************************************/
/******************************************************************************/

#include "Drawing.h"

/******************************************************************************/
/*  Definitions  */
/******************************************************************************/

static struct markedList {
	DrawingList user, found, hidden;
	Drawing super, class;
} MarkedLists;

/******************************************************************************/
/***  General routines for this module ***/
/******************************************************************************/

extern void init_mark ()
{
	MarkedLists.user   = Drawing_createList ();
	MarkedLists.found  = Drawing_createList ();
	MarkedLists.hidden = Drawing_createList ();
	MarkedLists.super  = NULL;
	MarkedLists.class  = NULL;
}

extern void finish_mark ()
{
	Drawing_freeList (MarkedLists.user);
	Drawing_freeList (MarkedLists.found);
	Drawing_freeList (MarkedLists.hidden);
	MarkedLists.super  = NULL;
	MarkedLists.class  = NULL;
}

extern void totally_unmark (o)
	Drawing o;
{
	/* !!! not quite right */
	remove_selection (o);
	free_found ();
/*	free_hidden (); !!! */
	if (o == MarkedLists.class)
		unmark_class (o);
	if (o == MarkedLists.super)
		unmark_supercomponent (o);
}


/******************************************************************************/
/* */
/******************************************************************************/

extern Drawing top_marked (list)
	MarkedList list;
{
	switch (list) {
	case m_user:
		return (first_drawingList (MarkedLists.user));
	case m_found:
		return (first_drawingList (MarkedLists.found));
	case m_supercomponent:
		return MarkedLists.super;
	case m_class:
		return MarkedLists.class;
	case m_hidden:
		return (first_drawingList (MarkedLists.hidden));
	default:
		sys_error ("Not defined for list type [%d]", list);
		return NULL;
	}
	/* NOTREACHED */
}

extern Cardinal numberOf_marked (list)
	MarkedList list;
{
	switch (list) {
	case m_user:
		return (sizeOf_drawingList (MarkedLists.user));
	case m_found:
		return (sizeOf_drawingList (MarkedLists.found));
	case m_hidden:
		return (sizeOf_drawingList (MarkedLists.hidden));
	case m_supercomponent:
	case m_class:
	default:
		sys_error ("Not defined for list type [%d]", list);
		return 0;
	}
	/* NOTREACHED */
}

extern void free_marked (list)
	MarkedList list;
{
	Drawing s;

	switch (list) {
	case m_user:
		FOR_DRAWINGLIST (s, MarkedLists.user)
			Drawing_deSelect (s);
		reset_drawingList (MarkedLists.user);
		break;
	case m_found:
		FOR_DRAWINGLIST (s, MarkedLists.found)
/*			Drawing_deEmphasize (s); */
		reset_drawingList (MarkedLists.found);
		break;
	case m_supercomponent:
		unmark_supercomponent (MarkedLists.super);
		break;
	case m_class:
		unmark_class (MarkedLists.class);
		break;
	case m_hidden:
	default:
		sys_error ("Not defined for list type [%d]", list);
	}
}

extern Drawing next_marked (list)
	MarkedList list;
{
	switch (list) {
	case m_user:
		return (next_drawingList (MarkedLists.user));
	case m_found:
		return (next_drawingList (MarkedLists.found));
	case m_hidden:
		return (next_drawingList (MarkedLists.hidden));
	case m_supercomponent:
	case m_class:
	default:
		sys_error ("Not defined for list type [%d]", list);
		return NULL;
	}
	/* NOTREACHED */
}

extern Drawing getN_marked (list, n)
	int n;
	MarkedList list;
{
	sys_error ("Not defined for any list type (yet)", list);
	return NULL;
}

extern Drawing second_marked (list)
	MarkedList list;
{
	switch (list) {
	case m_user:
		return (second_drawingList (MarkedLists.user));
	case m_found:
		return (second_drawingList (MarkedLists.found));
	case m_hidden:
		return (second_drawingList (MarkedLists.hidden));
	case m_supercomponent:
	case m_class:
	default:
		sys_error ("Not defined for list type [%d]", list);
		return NULL;
	}
	/* NOTREACHED */

}

extern void add_marked (list, o)
	MarkedList list;
	Drawing o;
{
	switch (list) {
	case m_user:
		if (Drawing_isSelected (o)) {
			return;
		} else {
/*			Drawing previous;
			previous = first_drawingList (MarkedLists.user);
			if (previous != NULL)
				Drawing_select (previous, select_normal);
			Drawing_select (o, select_first);
*/
			Drawing_select (o, select_normal);
			insert_drawingList (MarkedLists.user, o);
	        }
		break;
	case m_found:
/*		if (isEmphasized (o)) {
			return;
		} else {
			emphasize_drawing (o, show_emphasised);
			insert_drawingList (MarkedLists.found, o);
	        }
*/
		break;
	case m_supercomponent:
		/* !!! does o == MarkedLists.super ? */
		Drawing_select (o, select_super);
		MarkedLists.super = o;
		break;
	case m_class:
		/* Deselect previous selection */
		if (MarkedLists.class != NULL) {
			Drawing_deSelect (MarkedLists.class);
		}
		/*  Toggle chosen class */
		if (MarkedLists.class != o) {
			Drawing_select (o, select_normal);
			MarkedLists.class = o;
		}
		break;
	case m_hidden:
	default:
		sys_error ("Not defined for list type [%d]", list);
		return;
	}
}

extern void remove_marked (list, o)
	MarkedList list;
	Drawing o;
{
	switch (list) {
	case m_user:
		if (Drawing_isSelected (o) == False)
			return;
		Drawing_deSelect (o);
		extract_drawingList (MarkedLists.user, o);
		break;
	case m_found:
/*		if (isEmphasized (o) == False)
			return;
		deEmphasize_drawing (o);
		extract_drawingList (MarkedLists.found, o);
		if ((first = current_Selection ()) != NULL)
			Drawing_select (first, select_first);
*/
		break;
	case m_supercomponent:
		/* !!! does o == MarkedLists.super ? */
		Drawing_deSelect (MarkedLists.super);
		MarkedLists.super = NULL;
		break;
	case m_class:
 		/* !!! does o == MarkedLists.class ? */
		Drawing_deSelect (MarkedLists.class);
		MarkedLists.class = NULL;
		break;
	case m_hidden:
	default:
		sys_error ("Not defined for list type [%d]", list);
		return;
	}
}

extern void mark_drawingList (list, olist)
	MarkedList list;
	DrawingList olist;
{
	Drawing o;


	switch (list) {
	case m_user:
		FOR_DRAWINGLIST (o,  olist)
			add_selection (o);
		break;
	case m_found:
		FOR_DRAWINGLIST (o,  olist)
			mark_found (o);
		break;
	case m_supercomponent:
	case m_class:
	case m_hidden:
	default:
		sys_error ("Not defined for list type [%d]", list);
		break;
	}
}

extern void toggle_marked (list, o)
	MarkedList list;
	Drawing o;
{
	switch (list) {
	case m_user:
		if (Drawing_isSelected (o))
			remove_selection (o);
		else
			add_selection (o);
		break;
	case m_supercomponent:
		if (MarkedLists.super == o)
			unmark_supercomponent (o);
		else
			mark_supercomponent (o);
		break;
	case m_class:
		if (MarkedLists.class == o)
			unmark_class (o);
		else
			mark_class (o);
		break;
	case m_found:
	case m_hidden:
	default:
		sys_error ("Not defined for list type [%d]", list);
		break;
	}
}

/******************************************************************************/
/******************************************************************************/
/*  $Header: //gaia/group.local/computer/monet/lib/map/RCS/mark.c,v 1.1 1991/06/11 11:26:10 a38 Exp a38 $ */
/******************************************************************************/
/******************************************************************************/

