
/******************************************************************/
/******************************************************************/
/*  $Header: //gaia/group.local/computer/monet/lib/map/RCS/saveload.c,v 1.1 1991/07/22 11:36:41 a38 Exp a38 $  */
/******************************************************************/
/* This module is concerned with accessing map files, either from */
/* local files or from remote servers.                            */
/*                                                                */
/*  $Log: saveload.c,v $
 * Revision 1.1  1991/07/22  11:36:41  a38
 * Initial revision
 *
 * Revision 1.1  1991/06/11  11:26:10  a38
 * Initial revision
 *
*/
/*                                                                */
/******************************************************************/

#include "map/Drawing.h"
#include "map/Interface.h"
#include "general/MapServer.h"
#include "general/Shape.h"

#include <sys/types.h>
#include <sys/dir.h>

#define MAXMAPLINELEN 1024
#define MAPSPACEALLOCATION MAXMAPLINELEN * 16


/*****************************************************************************/
/* Change position of map */
/*****************************************************************************/

extern void Map_position(map, x, y, width, height)
	Map         map;
	Position        x, y;
	Dimension       width, height;
{
	Widget widget;
	Arg args[4];
/*
	if (width <= 0)
		width = 200;
	if (height <= 0)
		height = 200;
	widget = map_popup (map);
	if (widget == (Widget) 0) {
		sys_error ("No widget for this map");
		return;
	}
	debug_message ("Re positioning map %d %d %d %d", x,y,width,height);
	XtSetArg(args[0], XtNx, x);
	XtSetArg(args[1], XtNy, y);
	XtSetArg(args[2], XtNwidth, width);
	XtSetArg(args[3], XtNheight, height); 
	XtSetValues(widget, args, 4);
!!!	Map_adjust_size (map);
	Map_set_changed (map);
*/
}

/***************************************************************/
/*  Convert between String names and Enum values.              */
/***************************************************************/

struct {
	String          name;
	DrawnAs		drawn;
}               DrawnTable[] =
{
	{ "map",  drawnas_map },
	{ "listing", drawnas_list },
	{ NULL, drawnas_map }
};

extern String drawnas2string (d)
	DrawnAs d;
{
	int i;
	for (i=0; DrawnTable[i].name != NULL; i++)
		if (DrawnTable[i].drawn == d)
			return DrawnTable[i].name;
	return "sys_error";
}

extern DrawnAs string2drawnas (s)
	String s;
{
	int i;
	for (i = 0; DrawnTable[i].name != NULL; i++)
		if (EQS(s, DrawnTable[i].name))
			return DrawnTable[i].drawn;
	return drawnas_map;
}

/*****************************************************************************/
/*  Convert a map into a string, this format is suitable for transfer or     */
/*  saving.  Return NULL on error, caller must free the result */
/*****************************************************************************/

#define MAPATTRIBUTE(name, value)  \
	if (value && * value) { \
		(void) sprintf(result+len, "mapset %s %s\n", name, value); \
		len += strlen (result+len); \
	}

static Cardinal save_map_elements (map, result, len, available)
	Map map;
	String result;
	Cardinal len;
	Cardinal available ;
{
	Drawing d;

	(void) sprintf ((String) result+len, "#  Sub Map %s\n", Map_title (map));
	len += strlen ((String) result+len);
	(void) sprintf ((String) result+len, "map %s\n", Map_title (map));
	len += strlen ((String) result+len);
	MAPATTRIBUTE("domain", Map_get_domain (map));
	MAPATTRIBUTE("background_file", Map_get_background_file (map));
	MAPATTRIBUTE("background_colour", Map_get_background_colour (map));
	FOR_DRAWINGLIST(d, Map_elements (map)) {
		Position        x, y;
		Dimension       width, height;
		x = X(d);
		y = Y(d);
		width = WIDTH(d);
		height = HEIGHT(d);
		if (available - len < MAXMAPLINELEN) {
			available = (Cardinal) (len) + MAPSPACEALLOCATION;
			/*  !!! No recovery  */
			result = MonetRealloc (result, available);
		}
		(void) sprintf ((String) result+len, "ins %s %d %d %u %u\n", Drawing_title (d), x, y, width, height);
		len += strlen ((String) result+len);
		if (d->submap) {
			len = save_map_elements (d->submap, result, len, available);
			(void) sprintf ((String) result+len, "map %s\n", Map_title (map));
			len += strlen ((String) result+len);
		}
	}
	result[len] = NULL;
	return len;
}

extern String Map2String (map, result)
	Map map;
	String result;
{
	Cardinal len = 0;
	Cardinal available ;

	if (! result) {
		sys_error ("Result string is NULL");
		return NULL;
	}
	len = strlen (result);
	available = MAPSPACEALLOCATION + len;
	result = MonetRealloc (result, available);
	(void) sprintf(result+len, "# %s [%s]\n", pretty_time_stamp (), Map_title (map));
	len += strlen (result+len);
	(void) sprintf(result+len, "# This is a monetsh script");
	len += strlen (result+len);

	MAPATTRIBUTE("drawnas", drawnas2string(map->drawnas));
	MAPATTRIBUTE("font", Scale2String(map->scale));
	len = save_map_elements (map, result, len, available);
	return result;
}

#ifdef MONET_THIS_IS_NOT_NEEDED

/***************************************************************/
/*  Parse a map declaration.  These are found at the top of    */
/*  map files.  !!! Maybe this should be left to yacc.         */
/***************************************************************/

static Map parse_map(s)
	String          s;
{
	Boolean isclass;
	Entity superclass, result, entity;
	Map		map;
	char		*name;
	Scale		scale;
	char            expanded[1024], scale_name[1024], drawnas[1024], given_name[1024]; /* !!! */
	int             i, n;
	int		x, y;
	unsigned int	width, height;   /* !!! Not defined as Dimension, so sscanf will work */

	n = sscanf(s, "%s%d%d%u%u%s%s%s", given_name, &x, &y, &width, &height, expanded, scale_name, drawnas);
	debug_message ("Parse map [%s] [%s] %d %d %d %d [%s] [%s] [%s]", given_name, s, x,y,width,height,expanded,scale_name, drawnas);
	if (n != 8) {
		user_error("Can't parse map [%s]", s);
		return NULL;
	}
	entity = parse_title (given_name, &name, &superclass, &isclass);
	if (!entity && superclass == NULL) {
		user_error("Can't parse title [%s]", given_name);
		return NULL;
	}
	if (! isMapClass (superclass)) {
		user_error("Not a map [%s]", name);
		return NULL;
	}
	if (entity) {
		user_warning ("Map [%s] already exists", Entity_title (entity));
	} else if ((entity = Entity_create_instance (superclass, name)) == NULL) {
		user_error("Could not create [%s]", s);
		return NULL;
	}
/*	if (entity->map != NULL)
		map = entity->map;
	else if ((map = create_map (entity)) == NULL) {
		sys_error("Could not create map for [%s]", s);
		return NULL;
	}
*/
	Map_set_drawn_as (map, string2drawnas (drawnas));
	Map_position (map, (Position) x, (Position) y, (Dimension) width, (Dimension) height);

/*	if (strcmp(expanded, "expanded") == 0) {
		popup_map(map);
	} else if (strcmp(expanded, "closed") == 0) {
		popdown_map(map);
	} else {
		user_error ("Unknown expansion option [%s]", expanded);
	}
*/
	return map;
}

#endif /*  MONET_THIS_IS_NOT_NEEDED */

/*****************************************************************************/
/*  List local directory  */
/*****************************************************************************/

extern void Map_ListLocalDirectory (name)
	String name;
{
	(void) Monet_list_dir (name, "*:*");
}

/*****************************************************************************/
/*  remote directory list  */
/*****************************************************************************/

FileServerConnection Conn = NULL;

static Boolean list_dir (conn, code, message)
	String message;
	FileServerConnection conn;
	char code;
{
	
	switch (code) {
	default:
		user_error ("Unknown file server code [%c]", code);
		break;
	case fs_doing:
		writeln_report ("  %s", message);
		return True;
	case fs_end:
		writeln_report ("Listed remote directory");
		end_report ();
		break;
	case fs_error:
		user_error ("Error in transfer");
		break;
	case fs_locked:
		user_error ("Error in transfer, Locked");
		break;
	}
	Conn = NULL;
	return False;
}

static FILE * PullFile;

static Boolean pull_file (conn, code, message)
	String message;
	FileServerConnection conn;
	char code;
{
	switch (code) {
	default:
		user_error ("Unknown file server code [%c]", code);
		break;
	case fs_doing:
		fprintf (PullFile, "%s\n", message);
		return True;
		/* user_inform ("\t%s", message); */
		break;
	case fs_end:
		user_inform ("File transferred");
		break;
	case fs_error:
		user_error ("Error in transfer");
		break;
	case fs_locked:
		user_error ("Error in transfer, Locked");
		break;
	}
	fclose (PullFile);
	Conn = NULL;
	return False;
}

static Boolean save_file (conn, code, message)
	String message;
	FileServerConnection conn;
	char code;
{
	switch (code) {
	default:
		user_error ("Unknown file server code [%c]", code);
		break;
	case fs_doing:
		/* !!! This should never happen */
		user_inform ("\t%s", message);
		return True;;
	case fs_end:
		user_inform ("File saved on remote machine");
		break;
	case fs_error:
		user_error ("Error in transfer");
		break;
	case fs_locked:
		user_error ("Error in transfer, Locked");
		break;
	}
	Conn = NULL;
	return False;
}


extern Boolean Map_ListRemoteDirectory (host, port)
	String host;
	int port;
{
	if (port <= 0)
		port = Config_mapserver_port;
	if (host == NULL || *host == NULL)
		host = Config_mapserver_address;
	start_report ("Listing remote directory [%s+%d]", host, port);
	Conn = FileServer_request (host, port, list_dir, "dir");
	return (Conn != NULL);
}

extern Boolean save_remote_map (host, port, map)
	Map map;
	String host;
	int port;
{
	String request, user;

	user = Config_user_id;
	if (port <= 0)
		port = Config_mapserver_port;
	if (host == NULL || *host == NULL)
		host = Config_mapserver_address;

	request = MonetMalloc (64);
	(void) sprintf (request, "save %s %s\n", user, Map_title (map));
	request = Map2String (map, request);
	Conn = FileServer_request (host, port, save_file, request);
	MonetFree (request);
	return (Conn != NULL);
}


extern Boolean read_remote_map (host, port, name)
	String host;
	int port;
	String name;
{

	char  buff[1024]; /* !!! */

	if (port <= 0)
		port = Config_mapserver_port;
	if (host == NULL || *host == NULL)
		host = Config_mapserver_address;

	(void) sprintf (buff, "read %s", name);

	if (PullFile = MapServer_fopen_save (name)) {
		user_inform ("Pulling remote file [%s] from  [%s+%d]", name, host, port);
		Conn = FileServer_request (host, port, pull_file, buff);
		return True;
	}
	return False;
}

/*****************************************************************************/
/*****************************************************************************/

extern Map Map_get_map_for_entity (entity)
	Entity entity;
{
	Map m;
	FORALLMAPS (m)
		if (m->entity && m->entity == entity)
			return m;
	return NULL;
}


/***************************************************************/
/* Load a map, should be able to load from a file, or remotely */
/***************************************************************/

extern Boolean Map_load (map, filename)
	Map map;
	String filename;
{
	FILE *f;
	Entity class, entity;
	String title;
	char name [1024]; /* !!! */
	Boolean isclass;

	if (!map) {
		sys_error ("Map is NULL");
		return False;
	}
	if (! filename || ! *filename)
		filename = Map_title(map);

	(void) strcpy (name, MapServer_filename2name (filename));

	if ( ! Map_is_saved(map)) {
		user_error ("Map [%s] has not been saved", Map_title (map));
		return False;
	} else if (Map_classesOnly (map)) {
		user_error ("Already a class map");
		return False;
	} else if (entity = parse_title (name, &title, &class, &isclass)) {
		Map m;
		user_inform ("Entity found [%s]", Entity_title (entity));
		/*  Check to see if this map file is already loaded  */
		/*  Currently disallows this, could arrange for all  */
		/*  changes made to one copy are simultaneously made */
		/*  to the other.  */
		if (Map_get_map_for_entity (entity)) {
				user_error ("Map [%s] is already loaded", Entity_title (entity));
				return False;
			}
	} else if (class == NULL || title == NULL) {
		user_error("Can't parse title [%s]", name);
		return False;
	} else if (isclass) {
		user_error("Can't create map from class [%s]", name);
		return False;
	} else if ( ! isMapClass (class)) {
		user_error ("[%s] is not a map class", Entity_title (class));
		return False;
	}else if (!(entity = Entity_create_instance (class, title))) {
		user_error ("Could not create entity [%s:%s]", Entity_title (class), title);
		return False;
	}
	if (Map_is_loaded(map)) {
		user_inform ("Discarding Map");
		Map_disable_set_changed ();
		Map_discard (map);
		Map_enable_set_changed ();
		Map_set_changed (map);
		Map_request_redraw (map);
	}
	map->entity = entity;

	if (f = MapServer_fopen_load (filename)) {
		char name[1024], buffer[1024];   /* !!! */
		int x,y, n;
		unsigned int w,h;   /* These are defined as integers, not Position or Dimension
					   because sscanf works with integers */

		user_inform ("Loading map [%s]", Map_title (map));
		Map_disable_set_changed ();
		while (fgets(buffer, sizeof(buffer), f)) {
			Script_execute (buffer);
		}
		fclose(f);
		user_inform ("Loaded map [%s]", Map_title (map));
		Map_enable_set_changed ();
		Map_set_changed (map);
		Map_set_resized (map);
		Map_request_redraw (map);
		map->saved = True;
		return True;
	}
	user_inform ("New (empty) map created [%s]", Map_title (map));
	return False;
}

/***************************************************************/
/*  Save a map  */
/***************************************************************/

extern Boolean Map_save (map)
	Map map;
{
	FILE * f;
	String result;

	if (map->saved) {
		user_error ("Map is already saved");
		return False;
	}
	if (Map_classesOnly(map)) {
		user_error ("Can't save class map");
		return False;
	}
	result = MonetMalloc (64);
	if ( ! (result = Map2String (map, result))) {
		user_error ("Could not create string from map [%s]", Map_title (map));
		MonetFree (result);
		return False;
	}
	if (f = MapServer_fopen_save (Map_title(map))) {
		if (fputs (result, f) == 0) {
			user_error ("Error while saving map [%s]", Map_title (map));
		} else {
			user_inform ("Saved map [%s]", Drawing_title (map));
			map->saved = True;
		}
		fclose(f);
	}
	MonetFree (result);
	return False;
}

extern void map_server_init ()
{
	FileServer_init ();
}

extern void map_server_finish ()
{
	FileServer_init ();
}

/******************************************************************/
/******************************************************************/
/*  $Header: //gaia/group.local/computer/monet/lib/map/RCS/saveload.c,v 1.1 1991/07/22 11:36:41 a38 Exp a38 $  */
/******************************************************************/
/******************************************************************/

