/******************************************************************/
/******************************************************************/
/*  $Header: //gaia/group.local/computer/monet/lib/map/RCS/sizing.c,v 1.1 1991/07/16 11:08:23 a38 Exp a38 $  */
/******************************************************************/
/*  This is used for finding the correct sizes of drawings as they*/
/*  should appear on screen.  Because maps can be inside maps,    */
/*  drawings are not all drawn to the same scale, hence the need  */
/*  for this module.  */
/******************************************************************/

#include "map/Drawing.h"

static void	Scale2ActualPosition PR((Scale, Position, Position, Position*, Position*));
static void	ActualPosition2Scale PR((Scale, Position, Position, Position*, Position*));
static void	ActualSize2Scale PR((Scale, Dimension, Dimension, Dimension*, Dimension*));

extern void	Scale2ActualSize PR((Scale, Dimension, Dimension, Dimension*, Dimension*));
/******************************************************************/
/* Adjust position depending on scale of map */
/******************************************************************/

#define SCALE(a, offset, factor) (offset + a * factor)
#define DESCALE(a, offset, factor) ((-offset + a) / factor)

static void Scale2ActualPosition (scale, x,y, newx, newy)
	Scale scale;
	Position x,y, *newx, *newy;
{
	float fx,fy;
	Scale_xy (scale, &fx, &fy);
	*newx =  SCALE(x,0,fx);
	*newy =  SCALE(y,0,fy);
}

extern void Scale2ActualSize (scale, x,y, newx, newy)
	Scale scale;
	Dimension x,y, *newx, *newy;
{
	float fx,fy;
	Scale_xy (scale, &fx, &fy);
	*newx =  SCALE(x,0,fx);
	*newy =  SCALE(y,0,fy);
}

static void ActualPosition2Scale (scale, x,y, newx, newy)
	Scale scale;
	Position x,y, *newx, *newy;
{
	float fx,fy;
	Scale_xy (scale, &fx, &fy);
	*newx =  DESCALE(x,0,fx);
	*newy =  DESCALE(y,0,fy);
}

static void ActualSize2Scale (scale, x,y, newx, newy)
	Scale scale;
	Dimension x,y, *newx, *newy;
{
	float fx,fy;
	Scale_xy (scale, &fx, &fy);
	*newx =  DESCALE(x,0,fx);
	*newy =  DESCALE(y,0,fy);
}


/*****************************************************************/
/*  Find the position of a drawing that is marked  */
/*  Returns position relative to result map */
/*  and dimensions on absolute map  */
/*****************************************************************/

static Drawing map_get_marked (map, x_map, y_map, x_win, y_win, w_win, h_win)
	Map *map;
	Position *x_map, *y_map, *x_win, *y_win;
	Dimension *w_win, *h_win;
{
	Drawing s, d;

	FOR_DRAWINGLIST(s, Map_elements (*map))
		if (Drawing_is_marked (s, *x_map,*y_map)) {
			if (s->submap) {
				Position xl, yl, xr, yr;
				Map_corners (s->submap, &xl, &yl, &xr, &yr, False);
				if (xr && yr) {
					float w, h, factor;
					w = (float) (WIDTH(s)-2) / (float) xr;
					h = (float) (HEIGHT(s)-2) / (float) yr;
					if (w < h)
						factor = w;
					else
						factor = h;
					*x_map = (*x_map - X(s)-1) / factor;
					*y_map = (*y_map - Y(s)-1) / factor;
					*map = s->submap;
					if (d = map_get_marked (map, x_map,y_map, x_win, y_win, w_win, h_win)) {
						if (w_win && h_win) {
							*w_win *= factor;
							*h_win *= factor;
						}
						if (x_win && y_win) {
							*x_win = *x_win * factor + X(s) + 1;
							*y_win = *y_win * factor + Y(s) + 1;
						}
						return d;
                      			}
				}
			}
			if (w_win && h_win) {
				*w_win = WIDTH(s);
				*h_win = HEIGHT(s);
			}
			if (x_win && y_win) {
				*x_win = X(s);
				*y_win = Y(s);
			}

			return s;
		}
	return NULL;
}
/*****************************************************************/
/*  Takes window coordinates and returns a map they correspond   */
/*  with the coordinates for that map.  */
/*****************************************************************/

static Drawing map_actual2map (map, x_map, y_map, w_map, h_map)
	Map *map;
	Position *x_map, *y_map;
	Dimension *w_map, *h_map;
{
	Drawing s, d;
	Map original_map = *map;

	FOR_DRAWINGLIST(s, Map_elements (*map))
		if (Drawing_is_marked (s, *x_map,*y_map)) {
			if (s->submap) {
				Position xl, yl, xr, yr;
				Map_corners (s->submap, &xl, &yl, &xr, &yr, False);
				if (xr && yr) {
					float w, h, factor;
					w = (float) (WIDTH(s)-2) / (float) xr;
					h = (float) (HEIGHT(s)-2) / (float) yr;
					if (w < h)
						factor = w;
					else
						factor = h;
					*x_map = (*x_map - X(s)-1) / factor;
					*y_map = (*y_map - Y(s)-1) / factor;
					*map = s->submap;
					*w_map /= factor;
					*h_map /= factor;

					if (d = map_actual2map (map, x_map,y_map, w_map, h_map)) {
						if (d->submap)
							*map = original_map;
						return d;
                      			}
				}
			}
			return s ;
		}
	return NULL;
}

/*****************************************************************/
/*  Select a drawing or if none is marked lets the user create a */
/*  rectangle, every drawing inside will then be marked.         */
/*  Will select the resized drawing.                             */
/*****************************************************************/

static Boolean map_select_in_box (map, xloc, yloc, width, height)
	Position xloc, yloc;
	Dimension width, height;
	Map map;
{
	Drawing s, sub;
	Cardinal n = 0;

	FOR_DRAWINGLIST(s, map->elements)
		if (Drawing_in_box (s, xloc, yloc, width, height)) {
			sub = s;
			++n;
		}
	if (n == 1 && sub->submap &&
		xloc > X(sub) && yloc > Y(sub) &&
		xloc + width < X(sub) + WIDTH(sub) &&
		yloc + height < Y(sub) + HEIGHT(sub))
	{
		Position xl, yl, xr, yr;
		s = sub;
		Map_corners (s->submap, &xl, &yl, &xr, &yr, False);
		if (xr && yr) {
			float w, h, factor;
			w = (float) (WIDTH(s)-2) / (float) xr;
			h = (float) (HEIGHT(s)-2) / (float) yr;
			xloc = xloc - X(s)-1;
			yloc = yloc - Y(s)-1;
			if (w < h)
				factor = w;
			else
				factor = h;
			xloc /= factor;
			yloc /= factor;
			width  /= factor;
			height /= factor;
       			if (! map_select_in_box (sub->submap, xloc, yloc, width, height))
				toggle_selection (sub);
		} else
			toggle_selection (sub);
	} else if (n > 0) {
		FOR_DRAWINGLIST(s, map->elements)
			if (Drawing_in_box (s, xloc, yloc, width, height))
				if (Drawing_isClass (s))
					toggle_class (s);
				else
 					toggle_selection (s);
	} else
		return False;
	return True;
}

/******************************************************************/
/*  external routines  */
/******************************************************************/

extern Drawing Map_get_marked (map, x_map, y_map, x_win, y_win, w_win, h_win)
	Map *map;
	Position *x_map, *y_map, *x_win, *y_win;
	Dimension *w_win, *h_win;
{
	Drawing d;

	ActualPosition2Scale	((*map)->scale, *x_map, *y_map, x_map, y_map);
	d = map_get_marked	(map, x_map, y_map, x_win, y_win, w_win, h_win);
	if (x_win && y_win)
		Scale2ActualPosition	((*map)->scale, *x_win, *y_win, x_win, y_win);
	if (w_win && h_win)
		Scale2ActualSize	((*map)->scale, *w_win, *h_win, w_win, h_win);
	return d;
}

extern Drawing Map_actual2map (map, x_map, y_map, w_map, h_map)
	Map *map;
	Position *x_map, *y_map;
	Dimension *w_map, *h_map;
{
	Drawing d;

	ActualPosition2Scale ((*map)->scale, *x_map, *y_map, x_map, y_map);
	ActualSize2Scale ((*map)->scale, *w_map, *h_map, w_map, h_map);
	d = map_actual2map (map, x_map, y_map, w_map, h_map);

	return d;
}

extern Boolean Map_select_in_box (map, xloc, yloc, width, height)
	Position xloc, yloc;
	Dimension width, height;
	Map map;
{
	ActualPosition2Scale (map->scale, xloc, yloc, &xloc, &yloc);
	ActualSize2Scale (map->scale, width, height, & width, & height);
	return map_select_in_box (map, xloc, yloc, width, height);
}

/******************************************************************/
/******************************************************************/
/*  $Header: //gaia/group.local/computer/monet/lib/map/RCS/sizing.c,v 1.1 1991/07/16 11:08:23 a38 Exp a38 $  */
/******************************************************************/
/******************************************************************/

