/*****************************************************************************************/
/*****************************************************************************************/
/* $Header$ */
/*****************************************************************************************/
/*  This doesn't really make that much sense as yet because an snmp request will doesn't */
/*  return an up/down status, unless of course it is an interface.  Will deal with all   */
/*  this later.  */
/*****************************************************************************************/

#include <sys/types.h>
#include <netinet/in.h>
#include "snmp.h"
#include "query/SnmpLib.h"

/*****************************************************************************************/
/*  Globals  */
/*****************************************************************************************/

static Boolean Pinging;

struct snmp_query {
	char * host;
	char * community;
	char * variable;
	int len;
	int * data;
};
typedef struct snmp_query * SnmpQuery;
#define SNMP(m)  ((SnmpQuery)  SUBCLASSVALUE(m))
static QueryClass QP;

/*****************************************************************************************/
/*  Ping an snmp object  */
/*****************************************************************************************/

#ifdef CMU_SNMP

static Boolean request (query)
	Query query;
{
	Entity mob = query->entity;
	VStatus status = V_STATUS(mob);
	char *host = SNMP(query)->host, *community=SNMP(query)->community;
	struct snmp_session session, *ss;
	struct snmp_pdu *pdu, *response;
	struct variable_list *vars;
	oid name[MAX_NAME_LEN];
	int	arg;
	int	count, current_name = 0;
	char *names[128];
	int name_length;
	StatusMask mask = StatusNoneMask;


	bzero((char *)&session, sizeof(struct snmp_session));
	session.peername = host;
	session.community = (u_char *)community;
	session.community_len = MonetStrlen((char *)community);
	session.retries = SNMP_DEFAULT_RETRIES;
	session.timeout = SNMP_DEFAULT_TIMEOUT;
	session.authenticator = NULL;
	snmp_synch_setup(&session);
	if ((ss = snmp_open(&session)) == NULL) {;
		user_error ("Couldn't open snmp");
		SetReachStatusMask (status, status_unknown);
		return 0;
	}
	pdu = snmp_pdu_create(GET_REQ_MSG);

	add_oid2pdu(pdu, sysUpTime);
	add_oid2pdu(pdu, ifNumber);

	SetReachStatusMask (status, status_reachable);
	for (;;) {
		int snmpstatus;
		unsigned long uptime = 0;
		snmpstatus = snmp_synch_response(ss, pdu, &response);
		switch (snmpstatus) {
		case STAT_TIMEOUT:
			debug_message ("No Response from [%s]", host);
			SetReachStatusMask (status, status_unreachable);
			goto out;
		case STAT_ERROR:
		default:
			debug_message ("An error occurred");
			SetReachStatusMask (status, status_unknown);
			goto out;
		case STAT_SUCCESS:
			if (response->errstat == SNMP_ERR_NOERROR) {
				for(vars = response->variables; vars; vars = vars->next_variable) {
				IFOID(vars, sysUpTime)
					uptime = *vars->val.integer / 100;  /* sec / 100 */
				ELSIFOID(vars, ifNumber)
					SetConnectionsStatusMask (status, *vars->val.integer);
				ELSEOID
					debug_message ("Unknown snmp variable returned");
				ENDOID
				}
				goto out;
			} else {
				debug_message ("Error in packet. Reason: %s", snmp_errstring(response->errstat));
				if (response->errstat == SNMP_ERR_NOSUCHNAME) {
					printf("This name doesn't exist: ");
					for(count = 1, vars = response->variables; vars && count != response->errindex;
						vars = vars->next_variable, count++);
					if (vars)
						print_objid(vars->name, vars->name_length);
					printf("\n");
					SetReachStatusMask (status, status_invalid);
					goto out;
				}
				if ((pdu = snmp_fix_pdu(response, GET_REQ_MSG)) == NULL) {
					SetReachStatusMask (status, status_unknown);
					goto out;
				}
			}
			break;
		}
	}
	out:
	if (response)
		snmp_free_pdu(response);
	snmp_close(ss);

	return True;
}

#else /* CMU_SNMP */

static Boolean request (mob)
	Entity mob;
{
	return False;
}

#endif /* CMU_SNMP */



/********************************************************************/
/*  Free  */
/********************************************************************/

static void free (mob)
	Query mob;
{
	MonetFree (SNMP(mob)->host);
	MonetFree (SNMP(mob)->community);
	MonetFree (SNMP(mob));
}

/********************************************************************/
/*  Find  */
/********************************************************************/

static  Query find (text)
	String text;
{
	int cc;
	Query o;
	char host[512], community[512];

	cc = sscanf (text, "%[^ \t\n+]+%[^ \t\n]", community, host);
	if (cc != 2)
 		return NULL;

	FORINSTANCES (o, QP)
		if (EQS(SNMP(o)->community, community) && EQS(SNMP(o)->host, host))
			return o;

	return NULL;
}

/********************************************************************/
/*  Create  */
/********************************************************************/

static  SubValue create (query, text)
	String text;
	Query query;
{
	int cc;
	char host[512], community[512];

	cc = sscanf (text, "%[^ \t\n+]+%[^ \t\n]", community, host);
	if (cc == 2) {
		SnmpQuery new;
		new = MonetNew (struct snmp_query);
		new -> host      = MonetNewString (host);
		new -> community = MonetNewString (community);
		return (SubValue) new;
	} else {
		user_error ("%s:<community>+<host> seems to be incorrect [%s]", QP->title, text);
		return NULL;
	}
	/*  NOTREACHED */
}

/********************************************************************/
/*  Convert to text */
/********************************************************************/

static char * as_text (mob)
	Query mob;
{
	static char text [1024]; /* !!! */
	(void) sprintf (text, "%s:%s+%s", QP->title, SNMP(mob)->community, SNMP(mob)->host);
	return text;
}

/*****************************************************************************************/
/* Init and Finish */
/*****************************************************************************************/

static void finish ()
{
}

extern QueryClass init_snmp ()
{
	NEW_QUERYCLASS (QP, "snmp");
	if (snmplib_enabled ())
		QP->enabled  = True;
	else
		QP->enabled  = False;

	return QP;
}

/*****************************************************************************************/
/*****************************************************************************************/
/* $Header$ */
/*****************************************************************************************/
/*****************************************************************************************/

