/********************************************************************/
/********************************************************************/
/* $Header$ */
/********************************************************************/
/********************************************************************/

#include "general/General.h"
#include "x25.h"
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

/*****************************************************************************************/
/*  Definitions  */
/*****************************************************************************************/

struct x25_query {
	char * host;
};
typedef struct x25_query  * X25Query;
#define X25(m)  ((X25Query)  SUBCLASSVALUE(m))

static QueryClass QP = NULL;

/***************************************************************************/
/*  Load the resolver file  */
/***************************************************************************/


/* !!! Allocate dynamically, this lot reserves an megabyte */
#define MAXLOOKUP 4096
typedef struct lookup_x25 * LookupX25;

static struct lookup_x25 {
	char *host, * address;
	int addrlen;
	char * userdata;
	int userdatalen;
	LookupX25 next;
};

static LookupX25 Lookup;

static int LookupEntries = 0;
static time_t FileLoadedTime = 0;

/***************************************************************************/
/*  Read in a user data string definition and prodcue a byte string.       */
/***************************************************************************/

static void process_string (data, len)
	char * data;
	int *len;
{
	char special;
	char * c, *r;

	r = c = data;
	for ((*len)=0; *c!=NULL;c++, r++, (*len)++) {
		if (*len == 1023) {
			user_error ("Length of user data is greater than [%d]", 1024);
			break;
		}
		switch (*c) {
		case '\\':
			c++;
			switch (*c) {
			case 'n': *r = '\n'; break;
			case 'r': *r = '\r'; break;
			default:
				if (isdigit(*c)) {
					for (special= *c-'0',c++;isdigit(*c);c++)
						special = special * 10 + *c - '0';
					*r = special;
				} else
					*r = *c;
				break;
			}
		/*  Ignore double quotes, this is the NULL character  */
		case '\"': c++; break;
		default:
			*r = *c;
			break;
		}
	}
	*r = NULL;
}

static Boolean load_file (name)
	char * name;
{
	char host[1024], address[1024], buffer[1024], userdata[1024];
	int len;
	FILE * f;
	int cc;
	LookupX25 new;

	name = (char *) MonetFilename (name);
	f = fopen (name, "r");
	if (f == NULL) {
		user_error ("Cannot open X.25 resolver file [%s]", name);
		return False;
	} else {
		debug_message ("Loading x25 resolver file [%s]", name);
	}

	LookupEntries = 0;
	Lookup = NULL;
	while (fgets(buffer, sizeof(buffer), f) != NULL) {
		cc = sscanf (buffer, "%s %s %s", host, address, userdata);
		if (*host == '#') {
		} else if (cc >= 2) {
			int p;
			new = MonetNew (struct lookup_x25);
			new->host    = MonetNewString (host);
			new->address = MonetNewString (address);
			new->addrlen = MonetStrlen (address);
			new->next = Lookup;
			Lookup = new;
			if (cc == 2)
				strcpy (userdata, "\\1\\0\\0\\0");
			process_string (userdata, &(new->userdatalen));
			new->userdata = MonetMalloc (new->userdatalen);
			bcopy (userdata, new->userdata, new->userdatalen);
			if (cc != 2)
				debug_message ("[%s] user data len %d, data %x [%s]", host, new->userdatalen,
					(int) *(new->userdata), new->userdata);
			LookupEntries ++;
		} else if (cc != 0) {
			sys_error ("Error in  X25 host/address file");
		}
	}
	fclose (f);
	debug_message ("%d entries in the x25 translation table", LookupEntries);
	FileLoadedTime = time (0);
	return True;
}

static LookupX25 host2x121 (host)
	char * host;
{
	LookupX25 l;

	FORALL (l, Lookup, next)
		if (EQS(l->host, host))
			return l;

	return NULL;
}

extern void check_x25_file ()
{
	struct stat stats;
	LookupX25 l;
	char * name;

	name = (char *) MonetFilename (Config_x121_file);
	if (stat (name, &stats) != 0) {
		debug_message ("Error doing stat on [%s]", name);
		return;
	}
	if (stats.st_mtime > FileLoadedTime) {
		debug_message ("X25 file modified, reloading [%s]", name);
		/*  Should free previous entries first  */
		FORALL (l, Lookup, next) {
			MonetFree (l->host);
			MonetFree (l->address);
			MonetFree (l->userdata);
			MonetFree(l);
		}
		Lookup = NULL;
		load_file (Config_x121_file);
	}
}

/********************************************************************/
/*  Dummy pinger  */
/********************************************************************/

#if defined(GOULD_NP1) || defined (apollo) || defined (UBC_X25)
#else
static Boolean request (mob)
	Entity mob;
{
	/*  !!! Not implemented for this machine type  */
	return False;
}
#endif

/***************************************************************************/
/*  Pinger for UCB   !!! Untested */
/***************************************************************************/

#ifdef UBC_X25
#include <netccitt/x25_sockaddr.h>

static Boolean request (mob)
	Entity mob;
{
	int sd;
	struct x25_sockaddr address;
	
	if ((sd = socket (AF_CCITT, SOCK_STREAM, 0)) == NOTOK) {
		perror ("openx25: socket");
		return False; /* Error can be found in errno */
	} else if (connect (sd, &address, sizeof (address)) < 0) {
		perror ("openx25: connect");
		SetReachStatusMask (vstatus, status_unreachable);
	}
	close (sd);
	return False;
}
#endif

/***************************************************************************/
/*  the pinger for gould_np1  */
/***************************************************************************/

#ifdef GOULD_NP1

#include <stdio.h>
#include <errno.h>
#include <syslog.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/socketvar.h>
#include <netdb.h>
#include <netccitt/ccitt.h>
#include <netccitt/x25.h>
#include <netccitt/x25_user.h>


/* ====================================================================	*/
/* 	getcause(fd, error) - Diagnose error on openening a connection	*/
/* ====================================================================	*/
static void getcause (fd, error)
int             fd;
int             error;
{
	int             cause, diag, len;

	len = sizeof (int);

	if (getsockopt (fd, SOL_X25, X25_CAUSE, &cause, &len) < 0) {
		perror ("getsockopt X25_CAUSE");
		return;
	}

	if (getsockopt (fd, SOL_X25, X25_DIAG, &diag, &len) < 0) {
		perror ("getsockopt X25_DIAG");
		diag = 0;
	}

	debug_message (stderr, "Errno: %d, Cause: %d, diagnostics: %d\n",
	error, cause, diag);
}

/* ====================================================================	*/
/* 	openx25() - open a connection to address PSE_DTELOG		*/
/* ====================================================================	*/

static Boolean request (mob)
	Entity mob;
{
	VStatus vstatus = V_STATUS(mob);
	extern          		errno;
	register int    		sd;
	struct sockaddr_ccitt 		r_address;
	struct ccitt_addr 		ccitt_addr();
	int             		buffer_len;
	char            		buffer[128];
	LookupX25 lookup;

	if (! QP->enabled) {
		SetReachStatusMask (vstatus, status_unknown);
		return 0;
	}

	if ((lookup = host2x121 (X25(mob)->host)) == NULL) {
		SetReachStatusMask (vstatus, status_invalid);
		return 0;
	}
	r_address.scc_addr   = ccitt_addr(lookup->address);
	r_address.scc_family = AF_CCITT;
	if (r_address.scc_addr.s_length == -1) {
		debug_message ("address conversion failed");
		SetReachStatusMask (vstatus, status_invalid);
		return 0;
	}

	if ((sd = socket (AF_CCITT, SOCK_SEQPACKET, 0)) < 0) {
		perror ("ping_x25: socket");
		SetReachStatusMask (vstatus, status_unknown);
		return 0;
	} 

	bzero (buffer, sizeof(buffer));
	buffer_len  = lookup->userdatalen;
	bcopy (lookup->userdata, buffer, buffer_len);

	if (setsockopt (sd, SOL_X25, X25_CALLUSER, buffer, buffer_len) < 0) {
		perror ("openx25: setsockopt");
		SetReachStatusMask (vstatus, status_unknown);
	} else if (connect (sd, &r_address, sizeof (r_address)) < 0) {
		perror ("openx25: connect");
		SetReachStatusMask (vstatus, status_unreachable);
		getcause (sd, errno);
	} else {
		SetReachStatusMask (vstatus, status_reachable);
	}
	close (sd);
	return False;
}

#endif /* GOULD_NP1 */

/***************************************************************************/
/*  the pinger for apollos  */
/***************************************************************************/

#ifdef apollo

#include "/sys/ins/base.ins.c"
#include "/sys/ins/x25.ins.c" 

/* Macros to DEBUG info	*/

#define write_string(name,len)    {char * i=name; for(;i<(name+len);++i) putchar(*i); }
#define OUTSTATUS   printf ("Status = %x\n", status);
#define OUTS(name,len)  printf ("name:%d = [", len); write_string((char *)name,len); printf ("]\n");
#define OUTX(name) printf("name = %x\n", name);

static Boolean request (query)
	Query query;
{
	Entity mob = query->entity;
	VStatus vstatus = V_STATUS(mob);
	char	*Idadr = X25(query)->host;
	x25_$adr_t		x121addr ;
	short int		x121addrlen = 0;
	name_$name_t		hostname ;
	short int		hostnamelen = 0;
	name_$pname_t		gateway ;
	short int		gatewaylen = 0;
	status_$t		status, close_status;
	x25_$udata_t		userdata ;
	short int		userdatalen = 0;
	x25_$clear_data_t	clear;
	long int		handle = 0;
	char x121[1024], gate[1024], host[1024]; /* !!! */
	extern int errno;
	LookupX25 lookup;

	if (! QP->enabled) {
		SetReachStatusMask (vstatus, status_unknown);
		return 0;
	}

	if ((lookup = host2x121 (Idadr)) == NULL) {
		SetReachStatusMask (vstatus, status_invalid);
		return 0;
	}

	bzero (x121addr, sizeof(x121addr));
	bzero (userdata, sizeof(userdata));
	x121addrlen = lookup->addrlen;
	userdatalen = lookup->userdatalen;
	bcopy (lookup->address,  x121addr, x121addrlen);
	bcopy (lookup->userdata, userdata, userdatalen);

	/* Try to make the call */

	debug_message ("userdatalen= %d", userdatalen);
	x25_$call(gateway, gatewaylen, x121addr, x121addrlen, userdata, userdatalen, handle, clear, status);
	if (handle != 0) {
		x25_$close(handle, close_status);
		if (close_status.all != status_$ok) {
			sys_error ("x25_$close returned status [%lx]", close_status.all);
		}
	}

#ifdef DEBUG                                  
	printf("INFO after call setup :\n");
	OUTS(hostname, hostnamelen);
	OUTS(gateway, gatewaylen);
	OUTS(x121addr, x121addrlen);
	OUTSTATUS;

	(void) strncpy (gate, gateway, gatewaylen);
	gate[gatewaylen] = NULL;
	(void) strncpy (x121, x121addr, x121addrlen);
	x121[x121addrlen] = NULL;

	debug_message ("[%s][%s] Handle : %x , cause %x , diagnostic %x status %x", gate, x121, handle, (int) clear.cause, (int) clear.diagnostic, status);
#endif DEBUG


	/* See if call succeeded */

	if (status.all == status_$ok) {
		SetReachStatusMask (vstatus, status_reachable);
	} else if (status.all == x25_$link_not_up) {
		/*  !!! Should set a flag saying there is a problem and issue a warning */
		SetReachStatusMask (vstatus, status_unknown);
	} else {
		SetReachStatusMask (vstatus, status_unreachable);
	}
	return 0;
}
#endif /***** apollo *****/

/********************************************************************/
/*  Free  */
/********************************************************************/

static void free (mob)
	Query mob;
{
	MonetFree (X25(mob)->host);
	MonetFree (X25(mob));
}                           

/********************************************************************/
/*  Find  */
/********************************************************************/

static  Query find (host)
	String host;
{
	Query o;
	FORINSTANCES (o, QP)
		if (EQS(X25(o)->host, host))
			return o;
	return NULL;
}

/********************************************************************/
/*  Create  */
/********************************************************************/

static  SubValue create (query, host)
	Query query;
	String host;
{
	X25Query new;
	new = MonetNew (struct x25_query);
	new -> host = MonetNewString (host);
	return (SubValue) new;
}

/********************************************************************/
/*  Convert to name */
/********************************************************************/

static char * as_text (mob)
	Query mob;
{
	static char host [1024]; /* !!! */
	(void) sprintf (host, "x25:%s", X25(mob)->host);
	return host;
}

/********************************************************************/
/*  Init and finish  */
/********************************************************************/

static void finish ()
{
}

extern QueryClass init_x25 ()
{
	NEW_QUERYCLASS (QP, "x25");

#if defined(apollo)  || defined(GOULD_NP1)
	QP->enabled = load_file (Config_x121_file);   /*** !!! ***/
	if (QP->enabled == False) {
		user_error ("Sorry, X.25 monitoring disabled because of error in X25 Hosts File");
	}
#else
	QP->enabled = False;
	user_inform ("X.25 monitoring not implemented for this machine type");
#endif

	return QP;
}


/********************************************************************/
/********************************************************************/
/* $Header$ */
/********************************************************************/
/********************************************************************/


